#include<stdio.h>
#include<math.h>
#include<stdlib.h>
#include<string.h>
#include<time.h>


/*********Activate required Functions*******************/
//#define PARAMETRIC_STUDY			//Cases of parametric study | change SIZE_XINF
//#define GET_TRES_RND				//To sample random residence time from given RTD

#define LS				200.0		//Partition coeff of species S		|50 for PARAMETRIC_STUDY
#define DVES			2.0			//diameter of the vessel			|0.5	""
#define hL				2.0			//Height of LP						|0.5	""
#define hUNOM			0.10		//Nominal (initial) height of UP	|0.03	""
#define DIFFUSIVITY		4.803E-9 	//Diffusion of [S] in {Fe} at 1873 K

//End Point Criteria:
#define END_PT_FRAC		0.8			//fraction of equilibirum concentration acheived
#define TOTAL_TIME		1e5			//s

#define DELt			0.1
#define NUM_TIME		(int)(TOTAL_TIME/DELt)	//Estimate for number of time steps	
#define NUM_PRINT		50			//Number of time instances to be printed
#define NUM_DROP_TOT	1e7			//For malloc - total number of drops for the entirety
#define NUM_DROP_TRN 	1e6			//For malloc - number of drops in transition
#define NUM_BIRTH		1e5			//For malloc - number of drops born per time step

#define NUM_CASES_IN	9			//Number of Cases the Input Files have
#define NUM_CASES_RUN	9			//Number of Cases to Run

//Class limits for Distributions:
#define SIZE_XINF		0.04		//Maxmimum drop dia, m |0.02 for PARAMETRIC_STUDY

//Saturation Limit for drops:
#define SAT_LT			0.99		//Fraction of CSUeq; 90% and 99% takes ≈ same run time

#define PI				3.141592654
#define RANDNUM()		(rand()/((double)RAND_MAX)) // 0 < RANDNUM < 1
#define EQUAL(x, y) 	( fabs((x) - (y)) <= 1e-6*fabs(x) )
#define NORM(x, y)		pow((x)*(x) + (y)*(y), 0.5)		//L2 Norm
#define MIN(x, y)		(x > y ? y : x)
#define MAX(x, y)		(x > y ? x : y)
#define SIGN(x)			(((x) > 0)-((x) < 0))
#define r(x)			pow(x, 0.5)

typedef struct _size_tres_mtc_sampling {
	double *size_param, *meanTres_param, *meanTres_profile, *kdrop_profile,
		*dia, *Tres, *kdrop, kdropConst, entrainVol, *bin_size, Finf_size;
	int num_bin_size, num_drop;
} size_tres_mtc_sampling;

void size_tres_mtc_sampling_fn (size_tres_mtc_sampling *birth);
void drop_by_drop (FILE **fp, double *case_param, size_tres_mtc_sampling *birth);
void read_file_name(FILE *fp, char *filename);
int verify_input_columns(int N, FILE *fp);


int main()
{	
	FILE *fp[10];
	fp[0] = fopen("testing","w");
	fp[1] = fopen("input-size-distn-params","r");
	fp[2] = fopen("input-mean-res-times","r");
	fp[3] = fopen("input-mass-transfer-coeff-drops","r");
	fp[4] = fopen("input-case-params","r");
	fp[5] = fopen("concentration-profile","w");
	fp[6] = fopen("time-average-stats", "w");
	fp[7] = fopen("various-distribution-stats","w");
	fp[8] = fopen("size-distribution-stats","w");
	fp[9] = fopen("tres-distribution-stats","w");
	
	size_tres_mtc_sampling *birth = malloc(sizeof(*birth));
	
	birth->dia = malloc( NUM_BIRTH*sizeof(double *) );
	birth->Tres = malloc( NUM_BIRTH*sizeof(double *) );
	birth->kdrop = malloc( NUM_BIRTH*sizeof(double *) );

	int i, j, dummy, N;
	
#ifdef PARAMETRIC_STUDY
	
	int choice_size_distn_PS = 0;	//0, 1 or 2
	int choice_tres_distn_PS = 1;	//0, 1
	
	double size_param_PS[][3] = {
			2.0000, 0.7071, 0.7979,		/*Mixed*/
			0.5000, 1.4142, 0.3989,		/*Finer Drops*/
			4.0000, 0.7071, 0.7979		/*Coarser Drops*/
	};	
	double meanTres_param_PS[][3] = { 
			0, 30.0, 5.0,	
			0, 60.0, 5.0	/*Longer Tres*/
	};
	
	double case_param[][5] = {					
			//kflat, keye, kdrop, VPU, Xev			
	/*0*/	0.00050, 	0.002000, 	0.000500, 	0, 		0,
	/*1*/	0.00063, 	0.002519, 	0.000707, 	0, 		0.05,
	/*2*/	0.00072, 	0.002883, 	0.000866, 	0.1, 	0.1,
	/*3*/	0.00079, 	0.003173, 	0.001000, 	0.5, 	0.15,
	/*4*/	0.00085, 	0.003418, 	0.001118, 	1, 		0.2,
	/*5*/	0.00091, 	0.003632, 	0.001225, 	2, 		0.2,
	/*6*/	0.00096, 	0.003823, 	0.001323, 	4, 		0.2,
	/*7*/	0.00100, 	0.003997, 	0.001414, 	6, 		0.2,
	/*8*/	0.00104, 	0.004157, 	0.001500, 	8, 		0.2,
	/*9*/	0.00108, 	0.004306, 	0.001581, 	10, 	0.2,
	/*10*/	0.00111, 	0.004444, 	0.001658, 	12, 	0.2,
	/*11*/	0.00114, 	0.004575, 	0.001732, 	14, 	0.2,
	/*12*/	0.00117, 	0.004699, 	0.001803, 	16, 	0.2,
	/*13*/	0.00120, 	0.004816, 	0.001871, 	18, 	0.2,
	/*14*/	0.00123, 	0.004928, 	0.001936, 	20, 	0.2,
	/*15*/	0.00126, 	0.005035, 	0.002000, 	22, 	0.2,
	/*16*/	0.00128, 	0.005138, 	0.002062, 	24, 	0.2,
	/*17*/	0.00131, 	0.005236, 	0.002121, 	26, 	0.2,
	/*18*/	0.00133,	0.005332,	0.002179,	28,		0.2,
	/*19*/	0.00136,	0.005423,	0.002236,	30,		0.2
	};
	
#else
	/*****************Read Inputs from Files************************/
	//Input format: Each column must correspond to one 'case'
	
	double size_param[NUM_CASES_IN][5], meanTres_profile[NUM_CASES_IN][20],
		kdrop_profile[NUM_CASES_IN][20], bin_size[20];

	//Read Size Distribution Parameters (A, B, C) - drop diameter in mm:
	for(j=0; j<3; j++)
	for(i=0; i<NUM_CASES_IN; i++)
	dummy = fscanf(fp[1], "%lf ", &size_param[i][j]);
	
					
	//Read Mean Residence Times (s):
	N = 0;
	while (1)
	{
		if(!N && verify_input_columns(NUM_CASES_IN+1, fp[2])==0)
		return;	
		for(i=0; i<NUM_CASES_IN+1; i++)
		{
			//First column from input has droplet diameter in mm
			if(!i)	dummy = fscanf(fp[2], "%lf ", &bin_size[N]);
			else	dummy = fscanf(fp[2], "%lf ", &meanTres_profile[i-1][N]);
		}
		if(dummy > 0) N++;
		else break;
	}
	for(i=0; i<N; i++)
	bin_size[i] = bin_size[i]/1000;	//Convert droplet sizes to m
	
	birth->num_bin_size = N;
	birth->bin_size = bin_size;

	//Read Mass Transfer Coefficients for Drops (m/s):
	N = 0;
	while (1)
	{
		if(!N && verify_input_columns(NUM_CASES_IN, fp[3])==0)
		return;	
		for(i=0; i<NUM_CASES_IN; i++)
		dummy = fscanf(fp[3], "%lf ", &kdrop_profile[i][N]);
		if(dummy > 0) N++;
		else break;
	}
	
	//Read 'other parameters' - VPU, Aflat, Aeye, kflat, keye (SI units):
	double case_param[NUM_CASES_IN][5];
	for(j=0; j<5; j++)
	for(i=0; i<NUM_CASES_IN; i++)
	dummy = fscanf(fp[4], "%lf ", &case_param[i][j]);
					
#endif	//PARAMETRIC_STUDY


	double time_of_execution, A, B, C;
	clock_t begin_TOE = clock();

	for(i=0; i<NUM_CASES_RUN; i++)
	{		
		printf("\n\n\nCASE NUMBER - %d\n\n",i+1);		
		for(j=5; j<10; j++)
		fprintf(fp[j], "\n\n\nCASE NUMBER - %d\n\n",i+1);
	
#ifdef PARAMETRIC_STUDY		
		birth->size_param = size_param_PS[choice_size_distn_PS];
		birth->meanTres_param = meanTres_param_PS[choice_tres_distn_PS];	
		birth->kdropConst = case_param[i][2];											
#else	
		birth->size_param = size_param[i];
		birth->meanTres_profile = meanTres_profile[i];
		birth->kdrop_profile = kdrop_profile[i];
#endif	
		A = birth->size_param[0];
		B = birth->size_param[1];
		C = birth->size_param[2];
		birth->Finf_size = r(PI)/2*B*C*(1 + erf((log(SIZE_XINF*1000) - A)/B));	

		drop_by_drop (fp, case_param[i], birth);
		
		for(j=5; j<10; j++)
		fflush(fp[j]);
	}
	clock_t end_TOE = clock();
	time_of_execution = (double)(end_TOE - begin_TOE)/CLOCKS_PER_SEC;
	printf("\nCPU Time for running ALL CASES: %g\n", time_of_execution);

	//Finally Run the following:
	system("awk -f tabulate-c-vs-t-stats.awk < concentration-profile > tabulated-c-vs-t-stats");
	
	free(birth);
	for(i=0; i<10; i++)
	fclose(fp[i]);				
}


void drop_by_drop (FILE **fp, double *case_param, size_tres_mtc_sampling *birth)
{		
	int i, j, k, l, m, n;
	double kflat, keye, VPU, hU, VL, VU, VU0, entrainVol,
		 Aeye, Aflat, Aves, Reye, Rves, Xev;
	
	//Initialization is MUST to avoid re-use of previous function call values!
	hU = VL = VU = VU0 = Aeye = Aflat = Aves = Reye = VPU = kflat = keye = Xev = 0;
	
	Rves = DVES/2;
		
	Aves = PI*Rves*Rves;
	VL = PI*Rves*Rves*hL;
	VU = VU0 = PI*Rves*Rves*hUNOM;		
			
#ifdef PARAMETRIC_STUDY
	kflat = case_param[0];					
	keye = case_param[1];					
	VPU = case_param[3];
	Xev = case_param[4];
#else
	VPU = case_param[0];
	Aflat = case_param[1];					
	Aeye = case_param[2];					
	kflat = case_param[3];
	keye =  case_param[4];	
#endif
	entrainVol = VPU*VU/100*DELt;	//VU entrained per time step
		
	printf("kflat = %g \nkeye = %g \nVPU %g\n", kflat, keye, VPU);

		
	int N, M, noUP, Ndrop, step, num_time_drop, num_time;
	double t, d, CSU, CSL, CSLeq, CSUeq, CSdrop0, CSU0, CSL0, nU, sumdndrop,
		dnbulk, netNdrop, netAdrop, netVdrop, netkAdrop, SnetNdrop, SnetVdrop,
		SnetAdrop, SnetkAdrop, SAeye, SAflat, SkAflat, SkAeye,
		Xbulk, Xdrop, kAeff, TAnetArea, drop_contribution, NumEntRate;
						
	int *batch = malloc( (NUM_TIME+1)*sizeof(int *) );
	double *csl = malloc( (NUM_TIME+1)*sizeof(int *) ),
		*TAnetNdrop = malloc( (NUM_TIME+1)*sizeof(int *) ),
		*TAnetVdrop = malloc( (NUM_TIME+1)*sizeof(int *) ),
		*dia = malloc( NUM_DROP_TOT*sizeof(double *) ),
		*Tres = malloc( NUM_DROP_TOT*sizeof(double *) ),
		*Tdeath = malloc( NUM_DROP_TRN*sizeof(double *) ),
		*Adrop = malloc( NUM_DROP_TRN*sizeof(double *) ),
		*Vdrop = malloc( NUM_DROP_TRN*sizeof(double *) ),
		*kdrop = malloc( NUM_DROP_TRN*sizeof(double *) ),
		*CSdrop = malloc( NUM_DROP_TRN*sizeof(double *) );
	int *saturated = malloc( NUM_DROP_TRN*sizeof(double *) ),
		*alive = malloc( NUM_DROP_TRN*sizeof(double *) ),
		*swap = malloc( NUM_DROP_TRN*sizeof(double *) );				

	//Must initilialize to avoid hangover from previous Function calls:
	//works _as fast as_ the memset(array, 0, N*sizeof(array));
	for(i=0; i<(NUM_TIME+1); i++)
	batch[i] = csl[i] = TAnetNdrop[i] = TAnetVdrop[i] = 0;
	for(i=0; i<NUM_DROP_TOT; i++)
	dia[i] = Tres[i] = 0;
	for(i=0; i<NUM_DROP_TRN; i++)
	Tdeath[i] = Adrop[i] =  Vdrop[i] = kdrop[i] = \
		CSdrop[i] = saturated[i] = alive[i] = swap[i] = 0;

	CSL = CSL0 = 1.0;
	CSLeq = CSL/(LS*VU0/VL + 1);	//Equilibirum CSL, [S]eq
	CSUeq = CSLeq*LS;				//Equilibirum CSU, (S)eq
	noUP = N = step = M = Ndrop = nU = CSU = Xbulk = Xdrop = num_time_drop = 0;
	netNdrop = netAdrop = netVdrop = netkAdrop = SnetNdrop = SnetVdrop = 0;
	SnetAdrop = SnetkAdrop = SAeye = SAflat = SkAflat = SkAeye = NumEntRate = 0;
	
	double time_of_execution;
	clock_t begin_TOE = clock();
	
	//To monitor Time of Execution (TOE):
	double TOE_time_loop = 0, TOE_birth = 0, TOE_death = 0, TOE_mt_drops = 0;
	clock_t begin_time_loop, begin_birth, begin_death, begin_mt_drops;	
	
	/*******Time Loop********/
	begin_time_loop = clock();

	birth->entrainVol = 0;
	while( (CSL0 - CSL)/(CSL0 - CSLeq) < END_PT_FRAC  &&  step < NUM_TIME)
	{
		t = step*DELt;

#ifdef PARAMETRIC_STUDY
		if(step==0)
		hU = hUNOM/(1 - Xev*Xev);
		Reye = r(Rves*Rves - VU/(PI*hU));	//Transient, due to VU
		Aeye = 2*PI*Reye*hU;
		Aflat = Aves - PI*Reye*Reye;
#endif		
			
		//***BIRTH of Droplets
		begin_birth = clock();                    
		if(!noUP)
		{   
			birth->entrainVol += entrainVol;
					
			size_tres_mtc_sampling_fn (birth);		

			for(i=0; i<birth->num_drop; i++) 
			{
						 d	= birth->dia[i];
				kdrop[N] 	= birth->kdrop[i];		// Mass Transfer Coeff
				Tdeath[N]	= t + birth->Tres[i];	// Time of death
				Adrop[N] 	= PI*d*d;				// Area
				Vdrop[N] 	= PI/6.0*d*d*d;			// Volume 
				CSdrop[N] 	= CSU;		// its concentration (same as conc. of UP)
				saturated[N] = 0;
				alive[M++] 	= N;

				//Losses due to birth:
				VU = VU - Vdrop[N];          	//Volume of UP 
				nU = nU - Vdrop[N]*CSdrop[N];  	//Number of moles of S 
				
				//Transitional gains:
				netNdrop++;
				netVdrop  += Vdrop[N]; 			// Volume
				netAdrop  += Adrop[N]; 			// Area
				netkAdrop += kdrop[N]*Adrop[N];	// Capacity Coeffs created 
				
				if(Ndrop < NUM_DROP_TOT)		//For Statistics
				{
					dia[Ndrop] 	= d;
					Tres[Ndrop] = birth->Tres[i];
				}
				if(VU <= 0)
				{
					noUP = 1;		
					printf("\nUpper Phase completely drained after %d droplets"
							" at t = %g\n", Ndrop, t);
					break;
				}
				N++;
				Ndrop++;														
			}
			NumEntRate += batch[step] = birth->num_drop;
		}
		TOE_birth += (clock() - begin_birth);
		
		//***Mass Transfer at all Alive Droplets:
		begin_mt_drops = clock();
		 
		sumdndrop = 0;
		for(k=0; k<M; k++)
		{
			i = alive[k];
			if(!noUP && saturated[i])	continue;	
			
			CSdrop0 = CSdrop[i];	
			CSdrop[i] = LS*CSL + (CSdrop0 - LS*CSL)*exp(-kdrop[i]*Adrop[i]/(LS*Vdrop[i])*DELt);
			
			sumdndrop += (CSdrop[i] - CSdrop0)*Vdrop[i];	// To mass balance
			// When RHS becomes -ve, sumdndrop would return those moles to LP
			
			saturated[i] = (CSdrop[i] > SAT_LT*CSUeq);
			//this is to only cutoff mass trasfer of saturated drops;
			//droplets should not be rejoined until their Tdeath
		}
		TOE_mt_drops += (clock() - begin_mt_drops);

		//***DEATH of Droplets
		begin_death = clock(); 
		      
		if(!noUP)
		{                         
			m = 0;
			for(k=0; k<M; k++)
			{				
				i = alive[k];
				if(t > Tdeath[i])  
				{				
					//Gains due to rejoining drops
					VU += Vdrop[i];         		// Volume
					nU += Vdrop[i]*CSdrop[i]; 		// Number of moles of S
					
					//Transitional losses:
					netNdrop--;
					netVdrop  -= Vdrop[i];			// Volume
					netAdrop  -= Adrop[i];			// Area
					netkAdrop -= kdrop[i]*Adrop[i];	// Capacity Coeff
				}
				//Copy indices of Alive droplets
				else
				swap[m++] = i;
			}	
			M = m;
			//Rearrange ids of alive droplets
			for(k=0; k<M; k++)
			alive[k] = swap[k];
			
			//Restart arrrays when they are about to overflow:
			if(N >= NUM_DROP_TRN - NUM_BIRTH)
			{
				for(k=0; k<M; k++)
				{
					i = alive[k];
					Tdeath[k] 	= Tdeath[i];
					Adrop[k] 	= Adrop[i];
					Vdrop[k] 	= Vdrop[i];
					kdrop[k] 	= kdrop[i];
					CSdrop[k] 	= CSdrop[i];
					saturated[k] = saturated[i];
					alive[k] 	= k;
				}
				N = M;
			}
			num_time_drop = step;
		}
		TOE_death += (clock() - begin_death);
		
		//Even if UP drains the following must go on:
		SnetNdrop  += netNdrop;
		SnetVdrop  += netVdrop;
		SnetAdrop  += netAdrop;
		SnetkAdrop += netkAdrop;
		
		TAnetNdrop[step] = SnetNdrop/(step + 1);	//phi/step ≡ phi*dt/t
		TAnetVdrop[step] = SnetVdrop/(step + 1);
								
		//***Mass Transfer at UP = MT at (flat + eye) + Rejoining drops
		if(!noUP)  
		{     	
			CSU0 = CSU;
			CSU = LS*CSL + (CSU0 - LS*CSL)*exp(-(kflat*Aflat + keye*Aeye)/(LS*VU)*DELt );
			dnbulk = (CSU - CSU0)*VU;

			nU = nU + dnbulk;
			CSU = nU/VU;

			Xbulk = Xbulk + dnbulk;			
			Xdrop = Xdrop + sumdndrop;
			
			SAflat 	+= Aflat;
			SAeye 	+= Aeye;
			SkAflat += kflat*Aflat;
			SkAeye 	+= keye*Aeye;
		}
		else dnbulk = 0;
		
		//Conc of M in LP:
		csl[step] = CSL -= (dnbulk + sumdndrop)/VL;
		
		if(step==0)		printf("t  CSL  batch[step]  Ndrop	Pct UP Remaining\n");
		if(step%50==0)	printf("%g %g %d %d	%g\n", t, CSL, batch[step], Ndrop, VU/VU0*100);
				
		step++;
	} //End of time loop
	
	num_time = step;
	TOE_time_loop = (clock() - begin_time_loop);

	//Print profiles for selected time steps:
	double print_interval = MAX(1, (double)num_time/NUM_PRINT);
	for(j=0; j<NUM_PRINT; j++)
	{
		i = (int)(j*print_interval);
		if(i >= num_time) break;
		
		fprintf(fp[5],"%g %g\n", i*DELt, csl[i]);
		if(i <= num_time_drop)
		fprintf(fp[6],"%g %g %g\n", i*DELt, TAnetNdrop[i], TAnetVdrop[i]/VU0*100);
		//Execute the gnuplot script to view statistics	
	}

	//Why keff? To divide overall rate constant (obtained from model output) by netArea 
	//		and be able to compare with keff (obtained below)
	kAeff = (SkAflat + SkAeye + SnetkAdrop)/num_time;	//phi/total_step ≡ ∫phi*dt/T
	TAnetArea = (SAflat + SAeye + SnetAdrop)/num_time;
	drop_contribution = Xdrop/(Xbulk + Xdrop)*100;
	NumEntRate = NumEntRate/(num_time_drop*DELt);	//Overall number entrainment rate
		
	fprintf(fp[5],"\nPct-contribution-of-drops-to-mass-transfer %g\n", drop_contribution);
	fprintf(fp[5],"Steady-state-num-of-drops-entrained %d\n", (int)TAnetNdrop[num_time-1]);
	fprintf(fp[5],"Steady-state-pct-of-UP-entrained %g\n", TAnetVdrop[num_time-1]/VU0*100);
	fprintf(fp[5],"Effective-kA %g\n", kAeff);
	fprintf(fp[5],"Time-Avgd-Net-interfacial-area %g\n", TAnetArea);
	fprintf(fp[5],"Number-Entrainment-Rate: %g\n", NumEntRate);
	
	/******************************************************************************/
	/****Collect and Print Various Size, RTD and Entrainment Rate Statistics*******/	
	/******************************************************************************/	
	
	int num_bin_size = 21, num_bin_tres = 20, num_bin_Ndrop = 10;	//Number of bins
	int num_size[100] = {0}, num_tres[100][100] = {0}, num_Ndrop[100] = {0},\
		bin_Ndrop[100] = {0}, minNdrops = 123456789, maxNdrops = 0,\
		sum_num_tres[100] = {0}, to_skip_num_tres[100] = {0};
	double bin_size[100], bin_tres[100], sum_tres[100] = {0}, x, f, Sf, Sxf;	
	
	//Values for bins:	
	for(k=0; k<num_bin_size; k++)
	bin_size[k] = SIZE_XINF/20*k;
	for(i=0; i<num_bin_tres; i++)	
	bin_tres[i] = 0.5*(i+1);		//0.5 s interval	

	for(i=0; i<num_time; i++)	
	{
		minNdrops = MIN(minNdrops, batch[i]);
		maxNdrops = MAX(maxNdrops, batch[i]);
	}
	for(i=0; i<num_bin_Ndrop; i++)
	bin_Ndrop[i] = (minNdrops + i*(double)(maxNdrops - minNdrops)/num_bin_Ndrop);

	
	//Size and Residence Time
	Ndrop = MIN(Ndrop, NUM_DROP_TOT);
	for(i=0; i<Ndrop; i++)	
	{
		for(k=0; k<num_bin_size-1; k++)
		{
			if(dia[i] > bin_size[k] && dia[i] <= bin_size[k+1])
			{
				num_size[k+1]++;
				for(j=0; j<num_bin_tres-1; j++)
				{
					if(Tres[i] > bin_tres[j] && Tres[i] <= bin_tres[j+1])
					{	
						num_tres[k+1][j+1]++;
						break;	
					}
				}			
				break;
			}
		}
	}

	//Number-Entrainment-Rate:
	for(i=0; i<num_time; i++)
	for(j=0; j<num_bin_Ndrop-1; j++)
	{
		if(batch[i] > bin_Ndrop[j] && batch[i] <= bin_Ndrop[j+1])
		num_Ndrop[j+1]++;
	}
	
	//Print Statistics:
	fprintf(fp[7],"Size Distribution Statistics\n");
	for(k=0; k<num_bin_size-1; k++)
	{
		fprintf(fp[7], "%g %d\n", bin_size[k+1]*1000, num_size[k+1]);
		//For gnuplot:
		fprintf(fp[8], "%g %d\n", bin_size[k+1]*1000, num_size[k+1]);
	}
	fprintf(fp[7],"\nResidence Time Distribution Statistics\n");
	fprintf(fp[7], "d,mm--> ");
	for(k=0; k<num_bin_size-1; k++)
	fprintf(fp[7], "%g\t", bin_size[k+1]*1000);
	for(l=0; l<num_bin_tres-1; l++)
	{
		fprintf(fp[7], "\n%0.2f\t", bin_tres[l+1]);
		for(k=0; k<num_bin_size-1; k++)
		fprintf(fp[7], "%d\t", num_tres[k+1][l+1]);
	}

	fprintf(fp[7],"\n\nMean Residence Time Statistics\n");
	for(k=0; k<num_bin_size-1; k++)
	{
		Sf = Sxf = 0;
		for(l=0; l<num_bin_tres-1; l++)
		{
			x = bin_tres[l+1];
			Sf += f = num_tres[k+1][l+1];
			Sxf += x*f;
		}
		fprintf(fp[7], "%g	%g\n", bin_size[k+1]*1000, Sxf/Sf);
	}	
	
	fprintf(fp[7],"\n\nNumber-Entrianment Rate vs Their Frequency:\n");
	for(j=0; j<num_bin_Ndrop-1; j++)
	fprintf(fp[7],"%g %d\n", (double)bin_Ndrop[j+1]/DELt, num_Ndrop[j+1]);
	fprintf(fp[7],"Time-Average-Number-Entrainment Rate: %g\n", NumEntRate);
	
#ifdef GET_TRES_RND	
	//For gnuplot - Normalized Residence Time Distributions:
	for(k=0; k<num_bin_size-1; k++)	
	for(l=0; l<num_bin_tres-1; l++)
	{
		sum_num_tres[k+1] += num_tres[k+1][l+1];	
		if(num_tres[k+1][l+1]==0)
		to_skip_num_tres[k+1]++; 	//count how many zeros
	}
	for(k=0; k<num_bin_size-1; k++)
	{
		//Eliminate sparse distributions
		if(to_skip_num_tres[k+1] >= 15) //too many zeros!
		{
			fprintf(fp[9], "diameter = %g mm - SKIPPED\n", bin_size[k+1]*1000);
			continue;
		}
		fprintf(fp[9], "diameter = %g mm\n", bin_size[k+1]*1000);
		for(l=0; l<num_bin_tres-1; l++)
		fprintf(fp[9], "%g %g\n", bin_tres[l+1], (double)num_tres[k+1][l+1]/sum_num_tres[k+1]);
		fprintf(fp[9], "\n");
	}
#endif  //GET_TRES_RND

	//Time of Execution Statistics:
	TOE_time_loop = TOE_time_loop/CLOCKS_PER_SEC;	
	TOE_birth = TOE_birth/CLOCKS_PER_SEC;
	TOE_death = TOE_death/CLOCKS_PER_SEC;
	TOE_mt_drops = TOE_mt_drops/CLOCKS_PER_SEC;
	printf("\nTime of Execution Statistics:\n"
			"Time loop:		%g \n"
			"Birth of droplets:	%g (%0.0f pct)\n"
			"Death of droplets:	%g (%0.0f pct)\n"
			"Mass Transfer at drops:	%g (%0.0f pct)\n",\
			TOE_time_loop,\
			TOE_birth, TOE_birth/TOE_time_loop*100,\
			TOE_death, TOE_death/TOE_time_loop*100,\
			TOE_mt_drops, TOE_mt_drops/TOE_time_loop*100);
	
	free(batch);			
	free(dia);
	free(Tres);
	free(Tdeath);
	free(Adrop);
	free(Vdrop);
	free(kdrop);
	free(CSdrop);
	free(TAnetNdrop);
	free(TAnetVdrop);
	free(saturated);
	free(alive);
	free(swap);
	return;
}

void size_tres_mtc_sampling_fn (size_tres_mtc_sampling *birth)
{	
	/******************Droplet Sizes*******************/
	// from distribution, C/x*exp{ -[(ln(x) - A)/B]^2 }
	int i, j, N = 0;
	double A, B, C, Finf, z, d, erfinv, vol = 0, \
			a = 0.147, b = 4.3307, c;	//b = 2/(PI*a)

	A = birth->size_param[0];
	B = birth->size_param[1];
	C = birth->size_param[2];

	Finf = birth->Finf_size;

	while(vol < birth->entrainVol)
	{		
		z = 1.128379167*Finf*RANDNUM()/(B*C) - 1;		//1.128... = 1/r(PI/4)
		c = log(1 - z*z);
		erfinv = SIGN(z)*r( r( pow(b + c/2, 2) - c/a ) - (b + c/2) );
		
		birth->dia[N++] = d = exp(A + B*erfinv)/1000;		//diameter in m	

		vol += 0.5235987757*d*d*d;		//0.523... = PI/6		
	}
	birth->entrainVol -= vol;			//deduction
	birth->num_drop = N;
	

	//************Mean Residence Times and Mass Transfer Coefficients**************/
	
#ifdef PARAMETRIC_STUDY
	double meanTres, A1, B1, C1, SDTres, sigmean, URN;
	
	A1 = birth->meanTres_param[0];
	B1 = birth->meanTres_param[1];
	C1 = birth->meanTres_param[2]/1000;
	
	for(i=0; i<N; i++)	//Sample for all N droplets
	{
		meanTres = A1 + B1*exp(-(birth->dia[i])/C1);
		birth->Tres[i] = meanTres;
		birth->kdrop[i] = birth->kdropConst;
		
#ifdef GET_TRES_RND		//from Distribution, C/x*exp{ -[(ln(x) - A)/B]^2 }
		Finf = 1.0; 	//This means Xinf → infinity
		//whereas in size distn it was necessary to clamp to a finite size
		SDTres = 1.0;	//seems a good approximation ?
		
		sigmean = 1 + pow(SDTres/meanTres, 2);
		A = log( meanTres/r(sigmean) );
		B = 1.4142*r( log(sigmean) );			//1.4142 = r(2)
		C = 0.564189584/B;						//0.564... = 1.0/r(PI)
		
		z = 1.128379167*RANDNUM()/(B*C) - 1;	//1.128... = r(PI/4)
		c = log(1 - z*z);
		erfinv = SIGN(z)*r( r( pow(b + c/2, 2) - c/a ) - (b + c/2) );
		
		birth->Tres[i] = exp(A + B*erfinv);		//diameter in m
			
#endif	//GET_TRES_RND
	}
	
#else	

	int M = birth->num_bin_size, p1, p2, extrapolate;
	double x[M], y1, y2;
	
	for(i=0; i<M; i++)
	x[i] = birth->bin_size[i];	//droplet diameters
		
	//Find using Linear Interpolation and Extrapolation beyond bin range
	for(i=0; i<N; i++)	//For all new droplets born in this batch
	{
		d = birth->dia[i];
		if(d <= x[0])			//extrapolation
		{ 
			p1 = 0;	
			p2 = 1;
			extrapolate = 1;
		}
		else if(d > x[M-1]) 	//extrapolation
		{ 
			p1 = M-2;
			p2 = M-1;
			extrapolate = 1;
		}
		else {	
			extrapolate = 0;	//interpolation
			for(j=0; j<M-1; j++)
			{
				if( d > x[j] && d <= x[j+1] )
				{
					p1 = j;
					p2 = j+1;	
					break;
				}
			}
		}		
		y1 = birth->meanTres_profile[p1];
		y2 = birth->meanTres_profile[p2];
		birth->Tres[i] = y1 + (y2 - y1)*(d - x[p1])/(x[p2] - x[p1]);
		if(extrapolate)
		birth->Tres[i] = MAX(0, birth->Tres[i]);

		y1 = birth->kdrop_profile[p1];
		y2 = birth->kdrop_profile[p2];
		birth->kdrop[i] = y1 + (y2 - y1)*(d - x[p1])/(x[p2] - x[p1]);
		if(extrapolate)		//cap on 'k' is based on 2*r(D/PI/tc), tc = d_min/U_max
		birth->kdrop[i] = MIN(2*r(DIFFUSIVITY/(PI*0.0002/1.0)), birth->kdrop[i]);
	}	
#endif	//PARAMETRIC_STUDY		
}

void read_file_name(FILE *fp, char *filename)	//Copied from StackOverflow
{
    char path[1000], result[1000];
    int fd = fileno(fp); 
    /* Read out the link to our file descriptor. */
    sprintf(path, "/proc/self/fd/%d", fd);
    memset(result, 0, sizeof(result));
    readlink(path, result, sizeof(result)-1);
    strcpy(filename, result);
}
int verify_input_columns(int N, FILE *fp)			
{
	int i;
	char command[1000], filename[1000];
	read_file_name(fp, filename);
	strcpy(command, "awk '{if(NR==1) print NF}' <");
	strcat(command, filename);
	
	FILE *fp1 = popen(command, "r");
	int dummy = fscanf(fp1, "%d", &i);
	if(i==N) return 1;	//No issues		
	if(i < N)	printf("\nERROR: Missing Columns in the following File:");
	if(i > N)	printf("\nERROR: Excess Columns in the following File:");
	printf("\n%s\n\n", filename);
	return 0;
}
