#!/bin/bash

echo "Specific Gas Injection Rate (lpm/ton): (4, 2 or 1)"
read -r Qsp
if [ $Qsp -eq 4 ]
then
	width_factor="1.0"
elif [ $Qsp -eq 2 ]
then
	width_factor="1.0"
elif [ $Qsp -eq 1 ]
then
	width_factor="0.25"
else
	echo "Unknown value !!! Termniating !!!"
	echo 
	exit
fi

#Enter the GfsFunction | No spaces around "="
area_condition="(y - yIUL(x) < "$width_factor"*WIN )"
echo $area_condition
#The region to which interfacial area is measured covers all area that is NOT
#	1) the flat area
#	2) the droplets
# It cannot be argued that, since the portion of lip beyond certain depth (into LP)
# will be converted to droplets (in future time steps) they should not be accounted; 
# because, while the droplets are there for every time steps (statistical average), 
# the inclined interface to full depth is also alway there! (statistical average).
# So all interfacial area that is not 1) and 2) should be counted


# The 'escape sequence', "\" is to print quotes (")

for filename in zladle* 
do
	echo $filename
	awk -v var="$area_condition" 'BEGIN { to_print = 1; } {
			
 			if ( $1=="GfsTime") {
				i = $5;
				t = $8;
				print "GfsTime { i = "$5" t = "$8" end = "$8+0.00" dtmax = 0.01 }";
			}
			
			#Remove all lines in between the following:
 			if ($1=="GfsAdaptVorticity")
			to_print = 0;
			if ($1=="GfsPhysicalParams")
			to_print = 1;
						
			if(to_print==0)
			{			
				print "RemoveDroplets {start = end} F -1 (!S ? F : 0)";
				print "OutputView {start = end} image-%g.ppm {width = 500 height = 500} ULeye-z-0.gfv";
				print "OutputSimulation {start = end} zULeye-%g {variables = U,V,W,F}";	
								
				#// area("F") gives (interface_area/cell_vol) and OutputScalarSum is ∫ϕ.dV 
				#// First find areaF in Init and then use in OutputScalarSum 
				print "Init {istep = 1} { areaF = ("var" ? area(\"F\") : 0) }";
				print "OutputScalarSum {start = end} _ULeye_Area-%g {v = areaF}";
				
				print "GfsVariable inU";
				print "GfsVariable inV";
				print "GfsVariable inW";
				print "GfsVariable BcVelComp";
				print "OutputTime {istep = 1} GfsOutput-log";
				
				to_print = 2;
			}
			else if(to_print==1 && $1 != "GfsTime") 
			print $0;
					
		}' $filename > e-$filename	#'e' stands for edited
done


#To edit .gfv file with updated area_condition:

awk -v var="$area_condition" 'BEGIN {to_print = 1;} { 
	for(i=1; i<NF; i++)
	if($i=="return")
	{
		print "} { return "var"; } {";
		to_print = 0;
		break;
	}
	if(to_print)	print $0;
	to_print = 1;
	
}' < ULeye-z-0.gfv > tmp

awk '{print $0;}' < tmp > ULeye-z-0.gfv
rm tmp

echo "Press any key to exit"
read

# code for (batch processing) editing GfsRemoveDroplets line in the Gerris simulation file
# Go to terminal, type:  $ ./code-for-ULeye-Area.sh







