
// All operations are DIMENSIONLESS

#include<stdio.h>
#include<math.h>
#include<stdlib.h>

#define NUM_VAR		10		//1+Number of Input variables read from GfsOutput

//#define VERIFICATION		//To verify Nearest distance finding implementation
//#define ALL_SEARCH_ALL		//All volume nodes search for all VOF nodes

//Number of z-planes to be searched in +z and -z directions:
#define	NPz			3		//totally (2*NPz + 1) planes will be searched
							
/**** Choice of NPz (why NPz = 3 is good) *****
As NPz → ∞, nearest distances become same as that of ALL_SEARCH_ALL
But that is time consuming. So to optimize, try various NPz, plot freq. distn.
and compare those (using correlation coeff) with that of ALL_SEARCH_ALL 
NPz >= 3 gives a good correlation (R=0.991) with ALL_SEARCH_ALL,
provided all distances are outputted (w/o filtering) and compared 
*/


#define equal(x, y) ( fabs((x) - (y)) <= 1e-4*fabs(x) )
#define MIN(x, y)	(x > y ? y : x)
#define MAX(x, y)	(x > y ? x : y)
struct qsvmc { double s[NUM_VAR]; /*scalars*/};
void quick_sort (int var_id, struct qsvmc *v, int n);
void find_x_cutoff(int N, double xRef, struct qsvmc *v, double yCutoff, double *xCutoff);

main( int argc, char *argv[] )
{
	if(argc != 4) {
		printf("\nSome of the arguments to `dist-to-VOF.out' missing"
				"\n!!!Terminating!!!\n");
		return;
	}
	double dL = atof(argv[1]);			
	double dist_max = atof(argv[2]);
	double xRef = atof(argv[3]);

	FILE *fp1, *fp2, *fp3;
	fp1 = fopen("input", "r");
	fp2 = fopen("output", "w");
	fp3 = fopen("_Area_ULeye", "w");

	int i, j, k, N = pow(10, 7), M, m, n;
	double x, y, z, xCutoff, yCutoff;
	struct qsvmc *v = malloc( N*sizeof(struct qsvmc) );
		//vector: v[].s[0, 1, ...] = x, y, z, U, V, W, delV, VOFcond, areaF, dist;
	
#ifdef VERIFICATION
	dL = 0.1;
	dist_max = 0.5;
	xRef = +0.5;
	xCutoff = -0.2;
	yCutoff = -0.5;	
	
	//Generate Regular Grid with 'F' at x = 0
	N = 0;
	for(i=-5; i<=5; i++)
	for(j=-5; j<=5; j++)
	for(k=-5; k<=5; k++)
	{
		v[N].s[0] = x = i*dL;	//x
		v[N].s[1] = y = j*dL;	//y
		v[N].s[2] = k*dL;		//z
		v[N].s[3] = v[N].s[4] = v[N].s[5] = 0;	//U,V,W
		//F (VOF) is stepped-inclined line:
		v[N].s[7] = (x > 2*y/5-dL/2 && x < 2*y/5+dL/2 ? 0 : (x < 2*y/5 ? -1 : 1));	
		N++;
	}
	/*Uncomment to Test****
	for(i=0; i<N; i++) {
		for(j=0; j<7; j++) 
		printf("%g ", v[i].s[j]);
		printf("\n");
	}*/
	
#else

	//Step-1: Read Vectors:
	int dummy;
	N = 0;
	while (1)
	{
		for(i=0; i<NUM_VAR-1; i++)  
		dummy = fscanf(fp1, "%lf ", &v[N].s[i]);
		if(dummy > 0) N++;
		else break;
	}

	//Step-2: Find yCutoff - only above which the VOF nodes are considered
	//	to define xCutoff
	double yMIN = 123456789, yMAX = -123456789;
	for(i=0; i<N; i++)
	{
		yMIN = MIN(yMIN, v[i].s[1]);
		yMAX = MAX(yMAX, v[i].s[1]);
	}
	yCutoff = yMIN/3 + yMAX*2.0/3;

	//Step-3: Find xCutoff:
	find_x_cutoff(N, xRef, v, yCutoff, &xCutoff);
	printf("yCutoff: %0.4f\nxCutoff: %0.4f\n", yCutoff, xCutoff);
		
#endif
	//Step-4: Sort ascendingly w.r.t "z" coordinate:
	quick_sort(2, v, N);
	

	//Step-5: Collect ids of vectors of the VOF & Find Area_ULeye:
	int *id_VOF[10000], Nz, Nz_VOF[10000], NVOF;
	for (i=0; i<10000; i++) 
	id_VOF[i] = malloc( N*sizeof(double *) );	//2D array
	int *id_nearest = malloc( N*sizeof(int) );
	double z0 = v[0].s[2], Area_ULeye = 0;	
	Nz = k = NVOF = 0;
			
	for(i=0; i<N; i++)
	{
		if( (v[i].s[7] != -1 && v[i].s[7] != 1)	&&	/* VOF nodes */
			(v[i].s[0] >= xCutoff) )	//ULeye is to the right of xCutoff
		{			
			if(!equal(v[i].s[2], z0))	//New z-coordinate ? 
			{
				Nz++;					//New z-plane
				z0 = v[i].s[2];
				Nz_VOF[Nz-1] = k;		//Num of VOF nodes in a z plane 
				k = 0;
			}
			id_VOF[Nz][k++] = i;			//VOF ids
			NVOF++;						//Total number of VOF nodes
			
			Area_ULeye += v[i].s[8];
		}
	}
	Nz_VOF[Nz++] = k;
	
	fprintf(fp3, "%g %g", xCutoff, Area_ULeye);

	/*Uncomment to Test****
	for(j=0; j<Nz; j++)
	printf("%d\n", Nz_VOF[j]);
	for(j=0; j<Nz; j++)
	for(k=0; k<Nz_VOF[j]; k++) {
		i = id_VOF[j][k];
		fprintf(fp2, "%d %g %g %g\n", k, v[i].s[0], v[i].s[1], v[i].s[2]);
	}*/

	printf("\nNumber of All Vectors: %d\nNumber of VOF vectors: %d\n", N, NVOF);

	
	//Step-6: Find the nearest distance to VOF, for every node:
	//	Search is restricted within z-planes of range z+/-del
	int id, idz1, idz2, id_id;
	double dist1, dist2, pct = 0, del;

	m = n = 0;
	z0 = v[0].s[2];
	idz1 = 0;
	idz2 = NPz;
	del = dL;
	
	for(i=0; i<N; i++)	//Loop over all nodes
	{
		x = v[i].s[0];
		y = v[i].s[1];
		z = v[i].s[2];
			
		dist2 = 123456789;
		
#ifdef ALL_SEARCH_ALL /***************************************/
		for(j=0; j<Nz; j++)			//Loop over every z-plane
		for(k=0; k<Nz_VOF[j]; k++)	//VOF nodes	within a z-plane
		{
			id = id_VOF[j][k];
			
			//Compute for all N x M1 possiblities: 
			dist1 = pow(x - v[id].s[0], 2) + \
					pow(y - v[id].s[1], 2) + \
					pow(z - v[id].s[2], 2);
			dist2 = (dist2 > dist1 ? dist1 : dist2);	
			/*/For testing:
			if(dist2 > dist1) { dist2 = dist1; id_nearest[i] = id; }*/	
		}
#else /******************************************************/
	
		//search the matching z-plane
		m = 0;
		id = id_VOF[m++][0];
		while(fabs(z - v[id].s[2]) > 0.9*del && m < Nz)
		id = id_VOF[m++][0];
		
		//To search NPz planes in each of +z & -z directions:
		idz1 = MAX(0, m-NPz);
		idz2 = MIN(Nz-1, m+NPz);

		for(j=idz1; j<=idz2; j++)
		for(k=0; k<Nz_VOF[j]; k++)	//VOF nodes	within a z-plane
		{
			id = id_VOF[j][k];	
			
			dist1 = pow(x - v[id].s[0], 2) + \
					pow(y - v[id].s[1], 2) + \
					pow(z - v[id].s[2], 2);
			dist2 = (dist2 > dist1 ? dist1 : dist2);
			/*/For testing:
			if(dist2 > dist1) { dist2 = dist1; id_nearest[i] = id; }*/
			
		} //<<< Rate limiting step
		
#endif	//ALL_SEARCH_ALL
		
		v[i].s[9] = v[i].s[7]*pow(dist2, 0.5);	//signed distance
		n++;				
		/*/Task Monitor:
		if((double)(i+1)/N > pct)
		{
			if(i==0) printf("\nPercentage of Nearest Distances found:\n");
			printf("%g ", pct*100);
			fflush(stdout);
			pct += 0.1;
		}*/
	}
	printf("\nNearest Distances to VOF for %d Vectors Found\n\n", n);
	
	//Step-7: Sort according to nearest distance:
	quick_sort (9, v, N);
	
	//Step-8: Print: U,V,W,dist
	for(i=0; i<N; i++) 
	{
		if(fabs(v[i].s[9]) <= dist_max)		//Reject vectors greater than dist_max
		fprintf(fp2, "%g %g %g %g %g\n", \
					v[i].s[3], v[i].s[4], v[i].s[5], v[i].s[6], v[i].s[9]);			
		/*/For Testing
		id = id_nearest[i];	
		fprintf(fp2, "%g %g\n", v[i].s[0], v[ id ].s[1]);	*/
	}
	fclose(fp1);
	fclose(fp2);
	for (i=0; i<10000; i++) 
	free(id_VOF[i]);
	free(id_nearest);
	free(v);
}


	
void find_x_cutoff(int N, double xRef, struct qsvmc *v, double yCutoff, double *xCutoff)
{	
	/* x-cutoff location is
		-above which 'most' of the VOF cells of ULeye are located
		-this delineates the inclined UL from ULflat	*/
	
	int i, M = 0;
	double *x = malloc( 1e6*sizeof(double) );
	double mean = 0, SD = 0;	
	for(i=0; i<N; i++)
	{
		if( v[i].s[1] >= yCutoff &&
			v[i].s[7] == 0 )			//<< $VOF_condition

			mean += x[M++] = v[i].s[0];		//x coordinate
	}
	mean = mean/M;
	
	for(i=0; i<M; i++)
	SD = SD + pow(x[i] - mean, 2);
	SD = pow(SD/M, 0.5);

	*xCutoff = mean - SD/2;		//mean_x offset by SD_x away from xRef
	
	free(x);
}	

	
//Adapted from http://rosettacode.org/wiki/Sorting_algorithms/Quicksort#C
void quick_sort (int var_id, struct qsvmc *v, int n) 
{
	if (n < 2)
	return;
	struct qsvmc temp, *LEFT, *RIGHT;
	double pivot = v[n/2].s[var_id];	
	LEFT = v;			// address of LEFT most element
	RIGHT = v + n - 1;	// address of RIGHT most element
	while (LEFT <= RIGHT) 			
	{
		while (LEFT->s[var_id] < pivot)		LEFT++;
		while (RIGHT->s[var_id] > pivot)	RIGHT--;
		if (LEFT <= RIGHT) 
		{
			 temp = *LEFT;
			*LEFT++ = *RIGHT;
			*RIGHT-- =  temp;
		}
	}
	quick_sort(var_id, v, RIGHT-v+1);
	quick_sort(var_id, LEFT, v+n-LEFT);	
}	
	

