#!/bin/bash

#Some constants*********************/
#Cell Size at VOF in Gerris Simulation (dimensionless)
dL="0.01562"
#Maximum distance from interface considered for final statistics (dimensionless):
dist_max="0.5"
#Reference x-location (dimensionless): x-location of the inlet plane
xRef="0.5"
#xRef is used to find x-distances to VOF cells at ULeye which are then 
#	sorted in ascending to estimate the cutoff x-location. Only VOF cells 
#	above this cutoff are considered to be the inclined UL at eye.



#Step-0: Delete all backup files (files ending with ~, created by Gedit, emacs)

find . -type f -name '*~' -delete
wait


#Verify:
echo "Are _dist-to-VOF.out_ and _stats-near-VOF.out_"
echo "the latest versions ? (y/n) "
read -r ans
if [ $ans == "n" ]
then
	echo "exiting"
	exit
fi

			
#Step-1: Find Nearest Distances to VOF for all cells | also find Area_ULeye
rm ensemble
for filename in zVelocity-normal*
#for filename in zVelocity-normal-to-VOF-60
do
	echo $filename
	#Filter out all but the 'domain' nodes
	awk '{	if($1 != "#" && $4 > 0 ) { $4 = "";	print $0; }	}' < $filename > input
	wait
	
	#Run C program to Find Nearest Distances to VOF for all cells	
	echo $(./dist-to-VOF.out	$dL		$dist_max	$xRef)
	wait
	
	awk '{ print }' < output >> ensemble
	wait
	
	#Collect all ULeye-Area Values:
	awk '{ print }' < _Area_ULeye >> collect_Area_ULeye
	wait
done
rm _Area_ULeye

#Step-2: Calculate mean and Coeff of Variation of those Area_ULeye Values:		
awk 'BEGIN { N = 1; mean_x = 0; mean_A = 0; SD_x = 0; SD_A = 0; }
	{
		x[N] = $1;
		A[N++] = $2;
		mean_x = mean_x + $1;
		mean_A = mean_A + $2;
		print $0 > "Area_ULeye"
	} END {
		mean_x = mean_x/N;
		mean_A = mean_A/N;
		for(i=1; i<=N; i++)
		{
			SD_x = SD_x + (x[i] - mean_x)^2;
			SD_A = SD_A + (A[i] - mean_A)^2;
		}
		SD_x = ((SD_x/N)^0.5)/mean_x;
		SD_A = ((SD_A/N)^0.5)/mean_A;
		printf("Statistics xCutoff Area_ULeye\n");
		printf("Mean %g %g\n", mean_x, mean_A);
		printf("Coeff_of_Var %0.3f %0.3f\n", SD_x, SD_A);
	}' < collect_Area_ULeye > Area_ULeye
	rm collect_Area_ULeye
	

#Step-3: Sort vectors into bins according to their distance from VOF
echo $(./stats-near-VOF.out	$dL		$dist_max)
wait


#Step-4: Clean all ouptut simulation files
for filename in zVelocity-normal*
do
	rm $filename
done

rm ensemble input output GfsOutput-log


echo "Press any key to exit"
read

# Code for Batch Processing: 	Go to terminal, type:  $ ./code-to-run-gerris.sh


