#include<stdio.h>
#include<math.h>
#include<stdlib.h>

#define NUM_SCALAR		5

main( int argc, char *argv[] )
{
	if(argc != 3) {
		printf("\nSome of the arguments to `stats-near-VOF.out' missing"
				"\n!!!Terminating!!!\n");
		return;
	}
	double dL = atof(argv[1]);
	double dist_max = atof(argv[2]);
	
	FILE *fp1, *fp2;
	fp1 = fopen("ensemble", "r");
	fp2 = fopen("stats-output", "w");
	
	//Step-1: Read Vectors: U, V, W, dist	
	int i, j, k, N, dummy;
	double *v[NUM_SCALAR];
	for(i=0; i<NUM_SCALAR; i++)
	v[i] = malloc ( 1e7*sizeof(double) );	//2D array

	N = 0;
	while (1)
	{
		for(i=0; i<NUM_SCALAR; i++)  
		dummy = fscanf(fp1, "%lf ", &v[i][N]);
		if(dummy > 0) N++;
		else break;
	}
	
	//Step-2: Prepare coordinates for eta:
	//Dimensionless Distance coordinates
	int Nbin_UP =5, Nbin_LP = 7, Nbin = (Nbin_UP + 1 + Nbin_LP);
	double dist_nondim_UP[] = {-0.74, -0.45, -0.25, -0.11, -0.04};
	double dist_nondim_LP[] = {0.04, 0.11, 0.22, 0.35, 0.50, 0.70, 0.95};
			
	//Dimensional*** Distance coordinates for Bins
	//	(*** not the ultimate re-dimenionalization)
	double dist_bin[Nbin];
	
	for(j=0; j<Nbin_UP; j++)	//UP
	dist_bin[j] = dist_nondim_UP[j] * dist_max;
	
	dist_bin[Nbin_UP] = 0;		//VOF
	
	for(j=0; j<=Nbin_LP; j++)	//LP
	dist_bin[Nbin_UP + 1 + j] = dist_nondim_LP[j] * dist_max;
	
	//Step-3: Sort into Bins:
	int M = 1e7;	//Very large size required to avoid memory overlapping
	int num[Nbin], Nvec = 0;
	double *U[Nbin], *V[Nbin], *W[Nbin], *delV[Nbin];
	for(i=0; i<Nbin; i++)
	{
		U[i] =  malloc ( M*sizeof(double) );	
		V[i] =  malloc ( M*sizeof(double) );	
		W[i] =  malloc ( M*sizeof(double) );
		delV[i] =  malloc ( M*sizeof(double) );
	}
	double Uavg[Nbin], Vavg[Nbin], Wavg[Nbin], dVsum[Nbin],
			USD, VSD, WSD, dist, pct, dV, dV_sum;
	for(j=0; j<Nbin; j++)
	num[j] = Uavg[j] = Vavg[j] = Wavg[j] = dVsum[j] = 0;

	double del = 1.1*dL;
	for(i=0; i<N; i++)
	{
		dist = v[4][i];
		
		if(dist < 0)		// UP
		{
			for(j=0; j<Nbin_UP; j++)
			{
				//if(dist larger than lower limit but lesser than upper limit)
				if(dist >= dist_bin[j]-del && dist < dist_bin[j]+del)
				break;
			}
			if(j==Nbin_UP) continue;	//beyond range; rejected
		}
		
		else if(dist==0)	//VOF
		j = Nbin_UP;		
		
		else 				// LP
		{	
			for(j=Nbin_UP + 1; j<Nbin; j++)
			{
				if(dist >= dist_bin[j]-del && dist < dist_bin[j]+del)
				break;
			}
			if(j==Nbin) 	continue;	//beyond range; rejected
		}
		k = num[j];	
		U[j][k] = v[0][i];
		V[j][k] = v[1][i];
		W[j][k] = v[2][i];
		delV[j][k] = v[3][i];
		//For dV-weighted Average:	
		dVsum[j] += dV = v[3][i];
		Uavg[j] += v[0][i]*dV;
		Vavg[j] += v[1][i]*dV;
		Wavg[j] += v[2][i]*dV;
		num[j]++;
		Nvec++;
		
		/*/Task Monitor:
		if((double)(i+1)/N > pct)
		{
			if(i==0) printf("\nPercentage of Nearest Distances found:\n");
			printf("%g ", pct*100);
			fflush(stdout);
			pct += 0.1;
		}*/
	}
	
	//Standard Deviation and Print:
	for(j=0; j<Nbin; j++)
	{
		dV_sum = dVsum[j];

		//dV-weighted Averages:
		Uavg[j] = Uavg[j]/dV_sum;
		Vavg[j] = Vavg[j]/dV_sum;
		Wavg[j] = Wavg[j]/dV_sum;
		
		//dV-weighted Std Dev:
		USD = VSD = WSD = 0;
		for(k=0; k<num[j]; k++)
		{
			dV = delV[j][k];
			USD += pow(U[j][k] - Uavg[j], 2)*dV;
			VSD += pow(V[j][k] - Vavg[j], 2)*dV;
			WSD += pow(W[j][k] - Wavg[j], 2)*dV;
		}
		USD = pow(USD/dV_sum, 0.5);
		VSD = pow(VSD/dV_sum, 0.5);
		WSD = pow(WSD/dV_sum, 0.5);
		
		fprintf(fp2, "%0.3f %d %g %g %g %g %g %g\n",\
			 dist_bin[j], num[j], Uavg[j], Vavg[j], Wavg[j], USD, VSD, WSD);
	}
	printf("\nStatistics from %d vectors computed\n\n", Nvec);
	
	fclose(fp1);
	fclose(fp2);
	for (i=0; i<5; i++) 
	free(v[i]);
	for (i=0; i<Nbin; i++) 
	{
		free(U[i]);
		free(V[i]);
		free(W[i]);
		free(delV[i]);
	}
}


