#!/bin/bash


#Enter the GfsFunction | No spaces around "="

#Volume Condition specifies the domain (box volume) in which 
#	Velocity statistics have to be extracted around VOF
volume_condition="(box/2 - y < WIN) && (box/2 - y > WIN/3) && (box/2 - x < WIN*2)"
echo "Volume Condition is"
echo $volume_condition

#VOF Condition aims to delineate ULeye from ULflat
# 	 within $volume_condition, apart from specifying UP and LP.
#	Flags: [UP: -1] [LP:  1] [+x facing VOF cells: 0] [the rest of VOF: 123456789]
VOF_condition="(F==1 ? -1 : (F==0 ? 1 : (F_x > 0  && fabs(F_y) < 0.5 ? 0 : 123456789)))"
echo "VOF condition is"
echo $VOF_condition

#Verify:
#echo "Have you verified that this case is"
#echo "neither the case of Reduced Gravity nor Doubled rhoU ? (y/n)"
#read -r ans
#if [ $ans == "n" ]
#then
#	echo "change the _volume_condition_"
#	echo "Exiting"
#	exit
#fi

# The 'escape sequence', "\" is to print quotes (")

for filename in zladle* 
do
	echo $filename
	awk -v var="$volume_condition" -v var1="$VOF_condition" 'BEGIN { to_print = 1; } {
			
 			if ( $1=="GfsTime") {
				i = $5;
				t = $8;
				print "GfsTime { i = "$5" t = "$8" end = "$8+0.00" dtmax = 0.01 }";
			}
			
			#Remove all lines in between the following:
 			if ($1=="GfsAdaptVorticity")
			to_print = 0;
			if ($1=="GfsPhysicalParams")
			to_print = 1;
						
			if(to_print==0)
			{			
				#Actual Work goes here
				print "GfsInit {istep = 1} { domain = "var" }";
				print "GfsInit {istep = 1} { VOFcond = "var1" }";
					#// area("F") gives (interface_area/cell_vol)
				print "Init {istep = 1} { areaF = area(\"F\")*dV }";
				print "Init {istep = 1} { delV = dV }";
				print "OutputSimulation {start = end} zVelocity-normal-to-VOF-%g {"
				print "		variables = domain,U,V,W,delV,VOFcond,areaF format = text}";	
												
				#Dummy decclarations
				print "GfsVariable inU";
				print "GfsVariable inV";
				print "GfsVariable inW";
				print "GfsVariable BcVelComp";
				print "OutputTime {istep = 1} GfsOutput-log";
				to_print = 2;
			}
			else if(to_print==1 && $1 != "GfsTime") 
			print $0;
					
		}' $filename > e-$filename	#'e' stands for edited
done

wait


#Step-2: Run Gerris
for filename in e-zladle* 
do
	mpirun -np 8 gerris3D $filename
	wait
	rm $filename
done


echo "Press any key to exit"
read

# Go to terminal, type:  $ ./velocity-normal-to-VOF.sh







