#!/bin/bash

#Step-0: Use all Cores of system
num_cores=$(nproc)
echo
echo "This will run on "$num_cores" cores"

#Step-1: Find the Maximum of Number of records amongst all Trajectories:
awk_script=$(awk 'BEGIN {N0 = 0; M = 0;}
				 { 
				 	if($0 == '\n') 
				 	{ 
				 		N = NR - N0; 
				 		N0 = NR; 
				 		M = (M < N ? N : M); 
				 	} 
				 } END {print M + 100;}' <trajectory)
				 #Excess of 100 frames to avoid abrupt ending of movie


#Step-2: Run Gnuplot in Parallel:

#Read the strings in awk_script:
max_num_rec=($awk_script)
echo "Number of image frames to be produced: "$max_num_rec
echo

j=0
#For every "j" Gnuplot plots/outputs an image; each "j" is run on a CPU
while [ $j -lt $max_num_rec ]
do
	
	#Use of GNU Parallel: [syntax: $ parallel -j _num_of_jobs_ -- "cmd1" "cmd2" ... ]
	cmd_parallel="parallel -j $num_cores -- "
	
	#Generate $num_core images in Gnuplot simulatenously
	for ((i=0; i<$num_cores; i++))
	do
		((j++))
		cmd_gnuplot="gnuplot -e index=$j \"gnuplot-script-for-movie.gnu\""
		
		#concantenate the "cmd1" "cmd2" ... : [syntax:  a="$a cmd"]
		cmd_parallel="$cmd_parallel		\"eval ${cmd_gnuplot}\""
	done
	#echo ${cmd_parallel}
	eval ${cmd_parallel}
	
	echo -en "\r Number of frames produced: $j\c\b"
done


#Step-3: Stitch using ffmpeg

ffmpeg -r 60 -i "image_for_movie_%04d.png" -vcodec libx264 -vb 20M -r 25 movie.mp4
#		-r 60		Frame rate (Input) frames per second
#		-vb 1M		Video bit rate - 1 million bit/s 
#		-r 25		Frame rate (Output)

wait

#Step-4: Clean all the image files:
for filename in image_for_movie* 
do
	rm $filename
done

#//Go to terminal, type:  $ ./generate-movies.sh
