
#include<stdio.h>
#include<math.h>
#include<stdlib.h>
#include<string.h>
#include<time.h>


/*********Uncomment the following to activate that module*******************/
//#define 			CELLULAR	//Cellular Flow verification Cases | ! Deactivate BLEND !
//#define 			MASSLESS	//For inert tracer
//#define 			COARSE_BNDRY //To get a coarse solid boundary representation
#define 			BLEND		//For smooth transition in velocity jumps due to  new eddy
#define 			TOMIYAMA	//Drag Correlation | Comment out for Standard Drag Curve

//Integration Schemes:
//#define 			ANALYTICAL
#define 			IMPLICIT
//#define 			TRAPEZOIDAL


/***Ladle Input Parameters***/
#define Qinj		(int)4		//Specific Gas injection Rate
#define MAXLEV		(int)8
#define dL			(0.0097656*pow(2,(8-MAXLEV)))	//Cell size		*/

/*/For cellular flow verification***
#define Lcell		1.0			//cellular domain length
#define dL  		(Lcell/25)	//cell size		*/


/*********(r, z) of Particle at t = 0***********/
//(In LES, drops are in the centre of mainstream)

/*/	I: At the tip of Lip (axially) and in centre of mainstream (radially)
#define pr0			(Qinj==1 ? 0.27 : (Qinj==2 ? 0.40 : 0.54))		
#define pz0			(Qinj==1 ? 0.79 : (Qinj==2 ? 0.78 : 0.75))*/
//	II: At the start of Lip (axially) and in centre of mainstream (radially)
#define pr0			(Qinj==1 ? 0.27 : (Qinj==2 ? 0.38 : 0.51))		
#define pz0			(Qinj==1 ? 0.84 : (Qinj==2 ? 0.83 : 0.82))
#define pr0_range	0.0
#define pz0_range	0.0

#define T_TOT		500.0		//Total time
#define DT_MIN		2e-3		//Minimum time step size
#define DT_MAX		DT_MIN
#define NUM_PART	1000			//Number of Particles
#define NUM_PRINT	1000 		//Number of time instances to be printed
#define NUM_SCALAR	7			//Number of permanent scalars

#define g			9.81		//m/s^2
#define MUF			0.005		//Kg/m/s	Fluid Viscosity
#define SIGMA		1.0			//N/m		Interfacial Tension
#define RHOP		2800.0		//Kg/m^3	Particle density
#define RHOF		7000.0		//kg/m^3	Fluid density
	
//Probable size of row and column arrays:
#define NROW		(int)1000
#define NCOL		(int)1000

#define PI			3.141592654
#define RANDNUM()	(1.0-2.0*rand()/((double)RAND_MAX)) // -1 < RANDNUM < 1
#define EQUAL(x, y) ( fabs((x) - (y)) <= 1e-6*fabs(x) )
#define NORM(x, y)	pow((x)*(x) + (y)*(y), 0.5)		//L2 Norm
#define MIN(x, y)	(x > y ? y : x)
#define MAX(x, y)	(x > y ? x : y)
#define SIGN(x)		(int)(((x) > 0)-((x) < 0))
void find_ordinal (int num, char *ordinal)
{
	int x = (num <= 13 ? num : num%10);
	strcpy(ordinal, (x==1 ? "st" : (x==2 ? "nd" : (x==3 ? "rd" : "th"))));
	return;
}

struct nodal {
	int spoutBc, exist, BcType;
	double r, z, rIV, zIV, Ur, Uz, Vr, Vz, Ar, Az;
	double dUrdr, dUzdr, dUrdz, dUzdz;
	double TKE, eps, tE, tC, tD;
};
struct particle {
	int spoutBc;
	double r, z, Ur, Uz, Ar, Az, TKE, eps;
	double dUrdr, dUzdr, dUrdz, dUzdz;	
};
struct bpv { /*Boundary Position Vector*/
	double r, z;
	double rIV, zIV;	//boundary Inward Vector 
	int BcType;
};
struct nodal **make2Darray(int m, int n)
{
	int i;
	struct nodal **array;  
	array = malloc( m*sizeof(struct nodal*) );  
	for (i=0; i<m; i++)  
	array[i] = malloc( n*sizeof(struct nodal) ); 
	return array;
};
struct grid_params {
	 int NR, NC, NBN, *NRC;
	 double *rdist, *zdist;
	 struct nodal **node;
	 struct bpv *bp;
};
struct physical_properties { double rhoP, rhoF, Ucell, muF, sigma; };
struct qsvmc { double s[NUM_SCALAR+3]; /*scalars*/};
void quick_sort_multi_column ( struct qsvmc *v, int N, int M);
void quick_sort (int var_id, struct qsvmc *v, int n);
void concatenate_boundary_points (int N, struct bpv *p);
void reduce_boundary_edges (int *NBN0, struct bpv *bp, double angle, double interval);
double angle_bet_vectors(struct bpv *p);
int lagrangian_particle_tracking (FILE **fp, struct grid_params *grid, double RP, 
				 		struct physical_properties *prop, char *err_msg);
void bilinear_interpolation(int N, double *r, double *z, double S[N][4], double val[N]);						
void interpolate_scalars(int eddy, int NC, int NR, double *rdist, double *zdist,
		struct nodal **node, struct particle *p, double time, int particle_id, 
		double *TOE, FILE *fp);
int verify_boundary_conditions(int eddy, int NBN, struct particle *p, \
		struct bpv *bp, double time, int particle_id, double *TOE, FILE *fp);
void dimensionalize_vel_accel(int choice, int NC, int NR, 
							double Ucell, struct nodal **node);


main () 
{
	int i, j, k, l, n, M, N; 
	struct qsvmc v[30000];	//quick sort vector

	FILE *fp[10];
#ifdef CELLULAR
	fp[0] = fopen("input-cellular-field", "r");
	fp[1] = fopen("cellular-final-particle-position", "w");
	fp[2] = fopen("cellular-nondim-avg-rise-velocity", "w");	
#else
	char Q_case[4][10] = {"178lpm", "355lpm", "710lpm"};
	char input_file[50], boundary_file[50];
	sprintf (input_file, "scalars-%s-maxlev-%d",  Q_case [Qinj==1 ? 0 : (Qinj==2 ? 1 : 2)], MAXLEV);
	sprintf (boundary_file, "boundary-%s", Q_case [Qinj==1 ? 0 : (Qinj==2 ? 1 : 2)] );
	
	fp[0] = fopen(input_file, "r");
	fp[1] = fopen(boundary_file, "r");
	fp[2] = fopen("particle-radii", "r");
	fp[4] = fopen("domain-boundary", "w");
	fp[5] = fopen("trajectory", "w");
	fp[6] = fopen("residence-times", "w");
#endif
	fp[3] = fopen("velocity-field","w");
	fp[7] = fopen("testing", "w");

#ifdef CELLULAR
	printf("\nEnter choice (1 or 2) for Test/Simulation cases:\n"
			"1) Fluid Inertia Affected\n2) Flud Inertia Independent\t");	
	int case_choice;
	scanf("%d", &case_choice);
	if(case_choice > 2)
	{ printf("\nInvalid choice...Terminating\n\n");	return; }
#endif

	/*Step-1: Read Input from Gerris / Cellular Field, Arrange, sort columns *************/
	
	N = 0;
	int dummy;
	while (1)
	{
		for(i=0; i<NUM_SCALAR+3; i++)
		dummy = fscanf(fp[0], "%lf ", &v[N].s[i]);
		if(dummy > 0) N++;
		else break;
	}
	printf("\nTotal input vectors read %d\n", N);
			
	//Get rid of the first 3 columns 
	for(i=0; i<N; i++)
	for(j=0; j<NUM_SCALAR; j++)
	v[i].s[j] = v[i].s[j+3];

	//Sort Ascendingly w.r.t first 2 columns:
	quick_sort_multi_column (v, N, 2);		
	

	/*Step-2: Form Grid: Bin all scalars into node locations identified by 2D arrays*****/
			//node[ column ][ row ]		(row - fastest varying)
	int NRC[NCOL], NCR[NROW], NC, NR, num_nodes_filled = 0;
	double rdist[NCOL], zdist[NROW];

	//Step-2a: Collect Distinct rows (z) and columns (r):		
	struct qsvmc v1[30000];	//to sort 'z' values alone
	for(i=0; i<N; i++)
	v1[i].s[0] = v[i].s[1];
	quick_sort_multi_column (v1, N, 1);

	j = k = 0;
	rdist[0] = v[0].s[0] - dL;		
	zdist[0] = v1[0].s[0] - dL;
	for(i=0; i<N; i++)
	{
		if( !EQUAL(rdist[j], v[i].s[0]) )	{ j++;	rdist[j] = v[i].s[0];  }
		if( !EQUAL(zdist[k], v1[i].s[0]) )	{ k++;	zdist[k] = v1[i].s[0]; }
	}	
	rdist[j+1] = rdist[j] + dL;		
	zdist[k+1] = zdist[j] + dL;	
	NC = j+2;
	NR = k+2;
		
	//Verify if cell size (dL) at #define macro is precise:
	double dr_avg = 0;
	for(i=0; i<NC-1; i++)
	dr_avg += fabs(rdist[i+1] - rdist[i]);
	dr_avg = dr_avg/(NC-1);

	if( fabs(dL - dr_avg) > 1e-3*dr_avg ) {
		printf("\nCell Size at #define macro (dL = %g) is not Consistent "
				"with that of input grid (Average dr = %g)\n", dL, dr_avg);
		return;
	}
	
	struct nodal **node = make2Darray(NC+3, NR+3);
	struct nodal *nC, *nL, *nR, *nB, *nT, nDummy;	
	nDummy.exist = 0;
	
	for(j=0; j<NC; j++)
	for(k=0; k<NR; k++)
	node[j][k].exist = 0;	

	for(i=0; i<N; i++)
	v[i].s[NUM_SCALAR] = 0;

	//Step-2b: Match locations of input vector with that of (NC-2)x(NR-2) matrix:
	for(j=1; j<NC-1; j++)		//column (r)
	for(k=1; k<NR-1; k++)		//row (z)
	{
		for(i=0; i<N; i++)
		if( !v[i].s[NUM_SCALAR] &&
			 EQUAL(v[i].s[0], rdist[j]) && EQUAL(v[i].s[1], zdist[k]) )
		{
			nC = &node[j][k];
			nC->r 		= v[i].s[0];
			nC->z		= v[i].s[1];			
			nC->Ur 		= v[i].s[2];
			nC->Uz 		= v[i].s[3];		
			nC->TKE		= v[i].s[4];
			nC->eps		= v[i].s[5];
			nC->BcType 	= (int)v[i].s[6];		
			nC->exist 	= 1;
			num_nodes_filled++;
			v[i].s[NUM_SCALAR] = 1;	//to avoid re-visiting this vector
			break;
		}	
	}
	//Check if all input nodes have been filled in node[][]
	if(num_nodes_filled!=N)	{
		printf("\n %d of the input nodes not filled in node[][]\n", (N - num_nodes_filled));
		return;
	}				
	
	//Step-2c: Fill extra nodes (wrapper) around the boundaries:
	//		Scalars for wrapper nodes comes from neighbors in the direction
	// 		these nodes were added to the domain
	int nid;
	for(j=0; j<NC; j++)		//column (r)		
	for(k=0; k<NR; k++)		//row (z)
	{	
		if(node[j][k].exist)	continue;

		nC = &node[j][k];		//Current
		nL = &node[j-1][k];		//Left 
		nR = &node[j+1][k];		//Right
		nB = &node[j][k-1];		//Bottom
		nT = &node[j][k+1];		//Top	
		if(!j) nL = &nDummy;
		if(!k) nB = &nDummy;		
	
		nid = (nL->exist==1 ? 2 : (nR->exist==1 ? 3 :\
			  (nB->exist==1 ? 4 : (nT->exist==1 ? 5 : 0))));
		
		if(!nid) continue;
		
		//Copy all neighbor node information.. 
		*nC = (nid==2 ? *nL : (nid==3 ? *nR : (nid==4 ? *nB : *nT)));
		//...except node locations
		nC->r = (nid==2 ? nL->r+dL : (nid==3 ? nR->r-dL : nC->r));
		nC->z = (nid==4 ? nB->z+dL : (nid==5 ? nT->z-dL : nC->z));
		nC->exist = nid;	//won't be neighbor for any further non-existing nodes
	}
	printf("Scalars binned into respecitive nodes\n");

	//Visualization of Velocities using Gnuplot:
	double dt = 0.2;
	for(j=0; j<NC; j++)
	for(k=0; k<NR; k++)	
	{
		nC = &node[j][k];
		if(!nC->exist) continue;	
		fprintf(fp[3], "%g %g %g %g %g\n", nC->r, nC->z, nC->Ur*dt, nC->Uz*dt,\
						NORM(nC->Ur, nC->Uz));					
		//plot "velocity-field" using 1:2:3:4:5 every 1 with vectors linecolor palette z
	}
	
#ifndef CELLULAR	//!CELLULAR

	/*Step-3: Read and Form Domain Boundary**************************************/
	struct bpv bp[30000];
	N = 0;
	while (1)
	{
		for(i=0; i<5+3; i++)
		dummy = fscanf(fp[1], "%lf ", &v[N].s[i]);
		if(dummy > 0) N++;
		else break;
	}

	//Step-3b: Get rid of the first 3 columns 
	for(i=0; i<N; i++)
	for(j=0; j<5; j++)
	v[i].s[j] = v[i].s[j+3];

	//Step-3c: Sort Ascendingly w.r.t first 2 columns:
	quick_sort_multi_column (v, N, 2);	
	
	int NBN = 0;
	for(i=0; i<N; i++)
	{
		if(!v[i].s[4]) continue;
		bp[NBN].r = v[i].s[0];
		bp[NBN].z = v[i].s[1];
		bp[NBN].rIV = v[i].s[2];
		bp[NBN].zIV = v[i].s[3];
		bp[NBN].BcType = v[i].s[4];
		NBN++;
	}
	//Step-3d: Concatenate Boundary Points:
	concatenate_boundary_points (NBN, bp);		
	reduce_boundary_edges (&NBN, bp, 10, 10);

	for(i=0; i<=NBN; i++)	//For visualization at Gnuplot with vectors
	fprintf(fp[4], "%g %g %d %g %g\n", bp[i].r, bp[i].z, bp[i].BcType,\
							(bp[i].rIV - bp[i].r)*10, (bp[i].zIV - bp[i].z)*10);
	/*plot "domain-boundary" using 1:2:3 with line linecolor variable,\
	"domain-boundary" using 1:2:4:5 every 1 with vectors */
	printf("Domain Boundary Formed\n");	
	fflush(fp[4]);
	
#endif //!CELLULAR


	/*Step 4. Fluid Acceleration for the whole Eulerian Grid ******************/	
	for(j=1; j<NC-1; j++)
	for(k=1; k<NR-1; k++)
	{
		if(!node[j][k].exist) continue;
		
		nC = &node[j][k];		//Current
		nL = &node[j-1][k];		//Left 
		nR = &node[j+1][k];		//Right
		nB = &node[j][k-1];		//Bottom
		nT = &node[j][k+1];		//Top		
		if(!j)	{ nL = &nDummy; }
		if(!k)	{ nB = &nDummy; }
			  			
		//Fluid Acceleration (Verified using Analytical Velocity profile in 2D)
		
		//Gradient found from Central difference when two neighbor nodes exist
		// or Euler backward or forward when one of the neighbor nodes is missing
		
		//Note: In dXdr = (X_nR - X_nL)/(r_nR - r_nL), it is (X_nR - X_nL) 
		// and not (X_nL - X_nR) because r_nR > r_nL so that (r_nR - r_nL) = +2*dL

		nC->dUrdr = (nL->exist && nR->exist ? (nR->Ur - nL->Ur)/2 : \
							 	 (nR->exist ? (nR->Ur - nC->Ur) : (nC->Ur - nL->Ur)))/dL;
						
		nC->dUzdr = (nL->exist && nR->exist ? (nR->Uz - nL->Uz)/2 : \
							 	 (nR->exist ? (nR->Uz - nC->Uz) : (nC->Uz - nL->Uz)))/dL;
						
		nC->dUrdz = (nB->exist && nT->exist ? (nT->Ur - nB->Ur)/2 : \
							 	 (nT->exist ? (nT->Ur - nC->Ur) : (nC->Ur - nB->Ur)))/dL;
						
		nC->dUzdz = (nB->exist && nT->exist ? (nT->Uz - nB->Uz)/2 : \
							 	 (nT->exist ? (nT->Uz - nC->Uz) : (nC->Uz - nB->Uz)))/dL;
																	
		// dU/dt = Ur(∂U/∂r) + Uz(∂U/∂z), ∂U/∂t = 0
		nC->Ar = nC->Ur * nC->dUrdr + nC->Uz * nC->dUrdz;
		nC->Az = nC->Ur * nC->dUzdr + nC->Uz * nC->dUzdz;	
	}
	//Acceleration for Wrapper Nodes = Accel. of its corres. neighbor:
	for(j=0; j<NC; j++)
	for(k=0; k<NR; k++)
	{
		if(node[j][k].exist < 2)	continue;
		
		nC = &node[j][k];		//Current
		nL = &node[j-1][k];		//Left 
		nR = &node[j+1][k];		//Right
		nB = &node[j][k-1];		//Bottom
		nT = &node[j][k+1];		//Top	
			
		nC->Ar = (nC->exist==2 ? nL->Ar : (nC->exist==3 ? nR->Ar : \
				 (nC->exist==4 ? nB->Ar : nT->Ar)));
		nC->Az = (nC->exist==2 ? nL->Az : (nC->exist==3 ? nR->Az : \
				 (nC->exist==4 ? nB->Az : nT->Az)));
	}				 
	
	struct grid_params grid;
	grid.NR = NR;
	grid.NC = NC;
	grid.rdist = rdist;
	grid.zdist = zdist;
	grid.node = node;
#ifndef CELLULAR	//!CELLULAR
	grid.NBN = NBN;
	grid.bp = bp;
#endif		
	struct physical_properties prop;
		
#ifdef CELLULAR	//CELLULAR FLOW VERIFICATION:

	double Wcell, Qcell, Bcell, Rcell, gamma, Acell;
	printf("\nEnter value for \"R\"\n");
	scanf("%lf", &Rcell);
	Rcell = MAX(Rcell, 1e-10);
	Qcell = 1.25;
	Bcell = (10*PI);	// = 10 if Lcell = Lcell/PI
	gamma = 1.0/Rcell - 0.5;
	if(case_choice==1)	Acell = Rcell*Bcell/2;
	else				Acell = (2*PI); // = 2	if Lcell = Lcell/PI 
	
	double RP =	0.005;					//Particle radius, m	
	double VolP = 4.0/3*PI*pow(RP, 3);	//m^3
	prop.rhoP = 1.0; 					//kg/m^3
	prop.rhoF = prop.rhoP/gamma;
	double mP = VolP*prop.rhoP;			//kg
	double mF = VolP*prop.rhoF;	

	printf("\nCellular Flow Case Parameters: \ngamma = %g\nRcell = %g\n"
		"Acell = %g\nrhoF = %g\n",	gamma, Rcell, Acell, prop.rhoF);	
	
	double Wcell_values[] = { 0.01, 0.05, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 
						0.7, 0.8, 0.9, 1.0, 1.1, 1.2, 1.3, 1.4, 1.6, 1.8 };							
	int err;
	for(i=0; i<(case_choice==1 ? 1 : 18); i++)
	{
		Wcell = (case_choice==1 ? (1.0/Rcell - 1.5)*Qcell : Wcell_values[i]);
		prop.Ucell = pow( Lcell/(Acell*Wcell)*(mP - mF)*g/(mP + 0.5*mF),  0.5 ); //aka U0
		prop.muF = (mP - mF)*g/(6*PI*RP*prop.Ucell*Wcell);
		
		printf("Wcell = %g\nUcell = %g\nmuF = %g\n", Wcell, prop.Ucell, prop.muF);
		
		dimensionalize_vel_accel (1, NC, NR, prop.Ucell, node);
		/**************************************************************/
		err = lagrangian_particle_tracking (fp, &grid, RP, &prop, NULL);
		/**************************************************************/		
		dimensionalize_vel_accel (0, NC, NR, prop.Ucell, node);	//Revert
	}

#else	//LADLE ENTRAINMENT CASES
	int err;
	char err_msg[100];
	prop.rhoP	= RHOP;
	prop.rhoF	= RHOF;
	prop.muF	= MUF;
	prop.sigma 	= SIGMA;
	//Particle Radius in mm:
	double RP[100];
	N = 0;	
	while (1)
	{
		dummy = fscanf(fp[2], "%lf ", &RP[N]);
		if(dummy > 0) N++;
		else break;
	}
	printf("\nTotal number of particle sizes: %d\n", N);	
	
	for(i=0; i<N; i++)
	{
		fprintf(fp[6], "\n\nParticle_Radius(mm):  %g\n\n", RP[i]);
		fprintf(fp[7], "\n\nParticle_Radius(mm):  %g\n\n", RP[i]);		
		
		/****************************************************************/
		err = lagrangian_particle_tracking (fp, &grid, RP[i]/1000, &prop, err_msg);
		/****************************************************************/	
		if(err == -1) { printf("\n%s\n", err_msg); 	return;	}

		fflush(fp[6]);		
	}
#endif	//CELLULAR

	// Run "tabulate-rtd-statistics.awk":
	system("awk -f tabulate-rtd-statistics.awk < residence-times > tabulated-rtd-stats");
		
	for(i=0; i<10; i++)
	if(fp[i]) fclose(fp[i]);
				
	for (i=0; i<NC+3; i++)
	free(node[i]);  
	free(node);	
}	


int lagrangian_particle_tracking (FILE **fp, struct grid_params *grid, double RP, 
				 		struct physical_properties *prop, char *err_msg)
{		
	int NR = grid->NR;
	int NC = grid->NC;
	int NBN = grid->NBN;
	double *rdist = grid->rdist;
	double *zdist = grid->zdist;
	struct nodal **node = grid->node;
	struct bpv *bp = grid->bp;		
	double rhoF = prop->rhoF;	
	double rhoP = prop->rhoP;	
	double muF = prop->muF;
	double sigma = prop->sigma;

	double Uch = pow(2*g*RP, 0.5);	//some characteristic velocity for scaling
	double VolP = 4.0/3*PI*pow(RP, 3);	//Particle Volume
	double AreaP_spherical = PI*RP*RP;	//Projected area of spherical particle
	double mF = VolP*rhoF;			//mass of fluid
	double mP = VolP*rhoP;			//mass of particle
	double c1 = 0.5*mF;				//For Added Mass; 0.5 is added mass coefficient
	double c2 = rhoF*(2*RP)/muF;	//For particle Reynolds Number
	double c3 = 2*mP/rhoF;			//For particle relaxation time
	double Buoy = (mF - mP)*g;		//Buoyancy
	
#ifndef CELLULAR	//Ladle Entrainment
	
	double Mo = g*pow(muF, 4)*(rhoF - rhoP)/(pow(rhoF, 2)*pow(sigma, 3));	//Mortan Number
	double Eo = g*(rhoF - rhoP)*pow(RP*2, 2)/sigma;	//Eotvos Number
	double CD_Eo = 2.6667*(Eo/(Eo + 4));		//Drag coefficient
	
	printf("\n\nParticle Radius = %g mm\nEo = %g\nMo = %g\n\n", RP*1000, Eo, Mo);
			
#endif	//!CELLULAR
	
	//To monitor Time of Execution (TOE):
	double TOE_particle_loop = 0, TOE_time_loop = 0, TOE_interpolate = 0, TOE_verify_bc = 0;
	clock_t begin_particle_loop, begin_time_loop, begin_CD;	

	int i, j, k, l, M, N, NUM_RECORD = T_TOT/DT_MIN*10; 
	double *r = malloc( NUM_RECORD*sizeof(double *) );
	double *z = malloc( NUM_RECORD*sizeof(double *) );	
	double *rE = malloc( NUM_RECORD*sizeof(double *) );
	double *zE = malloc( NUM_RECORD*sizeof(double *) );
	double *lsE = malloc( NUM_RECORD*sizeof(double *) );
	double *time = malloc( NUM_RECORD*sizeof(double *) );	
	int num_out_stats = 9;
	double *stats[num_out_stats];
	for (i=0; i<num_out_stats; i++)  
	stats[i] = malloc( 1e8*sizeof(double *) );		
	
	char ordinal[3];
	int idcol, idrow, null_relUr, null_relUz, escape, new_eddy, split, N0, num_esc,\
		num_ensemble;				
	double t, dt, UFr, UFz, UPr, UPz, UPr0, UPz0, urms, tE, lE, tbE, 
		psir, psiz,	FrDrag, FzDrag, FrAM, FzAM, FrPG, FzPG, FrBuoy, FzBuoy, AreaP, \
		AreaPr, AreaPz, CDr, CDz, relUr, relUz, relUr0, relUz0, Rer, Rez, APr, APz,\
		tauRr, tauRz, expr, expz, Ar, Az, dr, dz, UPgradUr, UPgradUz, UAtauRr, UAtauRz,\
		TKEX, urmsX, UrX, UzX, psi_min = 10, psi_max = -10, psi_avg = 0, psir0, psiz0, \
		BFn_arg, BFn1, BFn2, num_split_steps, dt_split,	avg_rise_vel = 0, num_part_tot_time = 0,\
		t1r, t1z, sumdr = 0, sumdz = 0, Tres[NUM_PART], Tres_esc[NUM_PART], printdt; 
		
	struct particle p, eddy;
	struct nodal *nC;	
	double locr[] = {-0.8, -0.8, 0.8, 0.8};
	double locz[] = {0.8, -0.8, 0.8, -0.8};
	
	AreaPr = AreaPz = AreaP_spherical;	//as used in deriving CD correlations originally
		
	num_esc = M = urms = 0;
	begin_particle_loop = clock();
	
	for(l=0; l<NUM_PART; l++)	//particle loop
	//for(l=0; l<1; l++)
	{
		//Step 1: Particle Initialization:	
#ifdef CELLULAR			
		p.r = RANDNUM();
		p.z = RANDNUM();
#else		
		p.r = pr0 + pr0_range*RANDNUM();
		p.z = pz0 + pz0_range*RANDNUM();;
#endif		
		//p.r = locr[l];
		//p.z = locz[l];		
		UPr = UPz = N = escape = FrAM = FzAM = tauRr = tauRz = split = new_eddy = 0;
		BFn1 = 1;
		
		//Time loop for each particle:
		dt = DT_MIN;
		t = -dt;
		begin_time_loop = clock();
		while(t < T_TOT)
		{
			t += dt;				

			//printf("%g %g %g %g %g %g %g\n", t, p.r, p.z, UFr, UFz, UPr, UPz);
			
			//Step 2: Interpolate scalars at particle location p(r,z):
			interpolate_scalars(0, NC, NR, rdist, zdist, node, &p, t, l, &TOE_interpolate, fp[7]);
					
#ifdef MASSLESS
			p.r = p.r + p.Ur*dt;
			p.z = p.z + p.Uz*dt;
#else
#ifndef CELLULAR	//!CELLULAR
	
			//Step 3: Add Turbulence: Varied Eddy Interaction Model
			
			//Eddy Scales at particle location:
			urms = pow(0.6667*p.TKE, 0.5);
			tE = 2*0.201*p.TKE/p.eps;	
			lE = 2*0.164*pow(p.TKE, 1.5)/p.eps;

			if(t==0 && l==0)	//for monitoring stats at prX, pzX
			{ UrX = p.Ur;	UzX = p.Uz;		TKEX = p.TKE; 	urmsX = urms;}
					
			if( t==0 ||\
				t - tbE > tE ||\
				NORM(p.r - eddy.r, p.z - eddy.z) > lE ||\
				verify_boundary_conditions(1, NBN, &eddy, bp, t, l, &TOE_verify_bc, fp[7]) )
			{
				//New eddy:
				tbE = t;			//time of birth
				eddy.r = p.r;		//Location
				eddy.z = p.z;
				//Random number sampled	from Gaussian Distribution:
				//Note: the requirment is <psi> = 0 & <psi^2> = 1 
				psir = RANDNUM() + RANDNUM() + RANDNUM();	
				psiz = RANDNUM() + RANDNUM() + RANDNUM();	
				new_eddy = 1;
				
				if(t==0) 	//for monitoring stats at pr0, pz0
				{ 
					psi_avg += psir;				
					psi_min = MIN(psi_min, psir);
					psi_max = MAX(psi_max, psir);
				}
			}
			else	//Eddy motion due to local mean velocity
			{
				eddy.r += eddy.Ur*dt;	
				eddy.z += eddy.Uz*dt;	
				new_eddy = 0;
			}
						
			//Interpolate to find (Ur, Uz)_eddy for eddy transport in the next time step:
			interpolate_scalars(1, NC, NR, rdist, zdist, node, &eddy, t, l, &TOE_interpolate, fp[7]);

#endif	//!CELLULAR
#ifdef BLEND
			//To ensure smooth transition upon new eddy assignment:			
			num_split_steps = 10;
			if(new_eddy)
			{				
				dt = MAX(DT_MIN, dt)/num_split_steps;
				split = (new_eddy > 0);	
			}
			if(split > 0 && split <= num_split_steps)
			{
				//Blend_fn = 0.5 + 0.5*tanh[(X-A)/B]
				BFn_arg = (split - num_split_steps/2)/2;
				BFn1 = 0.5 + 0.5*tanh(BFn_arg);
				split++;
				dt_split = dt;
				N0 = N;
			}
			else
			{	
				split = 0;			// end of split steps
				BFn1 = 1;
				psir0 = psir;	
				psiz0 = psiz;
			}
			//Adaptive Time step size determined by tauR:
			if(!split)
			{
				//Smooth transition in time step after coming out of "split"
				BFn2 = 0.5 + 0.5*tanh((N-N0 - 3)/1);	//in 3*2=6 time steps
				
				dt = MAX(DT_MIN, MIN(DT_MAX, MIN(tauRr, tauRz)))*BFn2 + (1-BFn2)*dt_split;
			}
			
			//Instantaneous Fluid velocity known to particle:
			UFr = p.Ur + urms*(BFn1*psir + (1 - BFn1)*psir0);
			UFz = p.Uz + urms*(BFn1*psiz + (1 - BFn1)*psiz0);	
			//BFn ensures that velocity jump due to new eddy is gradual over time (albeit small)
#else
			//Adaptive Time step size determined by tauR:
			dt = MAX(DT_MIN, MIN(DT_MAX, MIN(tauRr, tauRz)));
		
			//Instantaneous Fluid velocity known to particle:	
			UFr = p.Ur + urms*psir;
			UFz = p.Uz + urms*psiz;			

#endif	//BLEND

			/***************************************************************************
			Step 4: Calculate Forces, Particle Acceleration, Velocity and Displacement:
			****************************************************************************/
			
			if(escape==2)	//attempted to cross impermeable boundary in previous step
			{ 
				UPr = UFr;		
				UPz = UFz; 
			}
			else 
			{
				UPgradUr = UPr*p.dUrdr + UPz*p.dUrdz;
				UPgradUz = UPr*p.dUzdr + UPz*p.dUzdz;
				//Gradual change of F_AM:
				FrAM = (FrAM + c1*(UPgradUr - APr))/2;
				FzAM = (FzAM + c1*(UPgradUz - APz))/2;	
			}
			FrPG = mF*p.Ar;
			FzPG = mF*p.Az;
			FrBuoy = 0;
			FzBuoy = Buoy*(1-p.spoutBc);		//No gravity at spoutBc
			
			relUr = fabs(UFr - UPr);
			relUz = fabs(UFz - UPz);
			null_relUr = (relUr < 1e-4*Uch);
			null_relUz = (relUz < 1e-4*Uch);
			relUr = MAX(relUr, 1e-6*Uch);
			relUz = MAX(relUz, 1e-6*Uch);
			
			//Reynolds number:
			if(escape==2)	//use old relU; including successive escapes
			{
				Rer = c2*relUr0;
				Rez = c2*relUz0;
			}
			else
			{
				Rer = c2*relUr;	
				Rez = c2*relUz;				
				relUr0 = relUr;
				relUz0 = relUz;			
			}	
			
			/*Drag Coefficients:*********************************************/		
#ifdef CELLULAR	//Stokes rising (creeping flow) = 24/Re
			CDr = 24.0/Rer;
			CDz = 24.0/Rez;
#elif defined(TOMIYAMA)	//1e-2 < Eo < 1e3; 1e-3 < Re < 1e6; 1e-14 < Mo < 1e7
			CDr = MAX( 24.0/Rer*(1 + 0.15*pow(Rer, 0.687)), CD_Eo);
			CDz = MAX( 24.0/Rez*(1 + 0.15*pow(Rez, 0.687)), CD_Eo);	
#else		//Standard drag curve:
			CDr = (Rer > 1000 ? 0.44 : 24.0/Rer*(1 + 0.15*pow(Rer, 0.687)));
			CDz = (Rez > 1000 ? 0.44 : 24.0/Rez*(1 + 0.15*pow(Rez, 0.687)));			
#endif	//ALL DRAG MODULES

			tauRr = c3/AreaPr/CDr/relUr;
			tauRz = c3/AreaPz/CDz/relUz;

			Ar = (FrAM + FrPG + FrBuoy)/mP;
			Az = (FzAM + FzPG + FzBuoy)/mP;	

#ifdef ANALYTICAL
			expr = exp(-dt/tauRr);
			expz = exp(-dt/tauRz);
			UAtauRr = UFr + Ar*tauRr;
			UAtauRz = UFz + Az*tauRz;		

			//Particle Velocity:
			UPr = null_relUr ? UPr + Ar*dt : UAtauRr - (UAtauRr - UPr)*expr;
			UPz = null_relUz ? UPz + Az*dt : UAtauRz - (UAtauRz - UPz)*expz;	
			
			//Particle Displacement:
			dr = UAtauRr*dt - tauRr*(UAtauRr - UPr)*(1-expr);
			dz = UAtauRz*dt - tauRz*(UAtauRz - UPz)*(1-expz);
			dr = (null_relUr ? UPr*dt - Ar*dt*dt/2 : dr);
			dz = (null_relUz ? UPz*dt - Az*dt*dt/2 : dz);
			//relU==0 ? dx = UP0*dt + A*dt²/2 = (UP-A*dt)*dt + A*dt²/2		

#elif defined(IMPLICIT)
			//Particle Velocity:
			UPr0 = UPr;
			UPz0 = UPz;
			UPr = null_relUr ? UPr + Ar*dt : (UPr + dt*(Ar + UFr/tauRr))/(1 + dt/tauRr);
			UPz = null_relUz ? UPz + Az*dt : (UPz + dt*(Az + UFz/tauRz))/(1 + dt/tauRz);
			//Particle Displacement:
			dr = 0.5*dt*(UPr0 + UPr);
			dz = 0.5*dt*(UPz0 + UPz);

#elif defined(TRAPEZOIDAL)		
			//Particle Velocity:
			UPr0 = UPr;
			UPz0 = UPz;
			t1r = dt/tauRr;
			t1z = dt/tauRz;
			UPr = (null_relUr ? UPr + Ar*dt : \
						(UPr*(1-0.5*t1r) + t1r*(UFr + 0.5*dt*UPgradUr) + dt*Ar)/(1 + 0.5*t1r));
			UPz = (null_relUz ? UPz + Az*dt : \
						(UPz*(1-0.5*t1z) + t1z*(UFz + 0.5*dt*UPgradUz) + dt*Az)/(1 + 0.5*t1z));
			//Particle Displacement:
			dr = 0.5*dt*(UPr0 + UPr);
			dz = 0.5*dt*(UPz0 + UPz);

#endif	//ALL INTEGRATION SCHEMES
			
			p.r += dr;
			p.z += dz;	
			sumdr += fabs(dr);
			sumdz += fabs(dz);
	
			//Particle Acceleration (for Added Mass in next time step):	
			tauRr = c3/AreaPr/CDr/relUr;
			tauRz = c3/AreaPz/CDz/relUz;
			APr = !t ? Ar : (APr + ((UFr - UPr)/tauRr + Ar))/2;	
			APz = !t ? Az : (APz + ((UFz - UPz)/tauRz + Az))/2;			
			//Accelerations are averaged to avoid their destabilzing oscillations
			//Here it is equivalent to using the central difference value of dU/dt
			//  of 'previous time step'; ie AP(i) ⟵ AP(i-1) = [U(i-2) - U(i)]/(2∆t)

#endif	//MASSLESS			
#ifdef CELLULAR
			//Step 5: Apply Periodic BC:
			if(p.r < -Lcell)	p.r = +Lcell + (p.r + Lcell);	
			if(p.r > +Lcell)	p.r = -Lcell + (p.r - Lcell);
			if(p.z < -Lcell)	p.z = +Lcell + (p.z + Lcell);
			if(p.z > +Lcell)	p.z = -Lcell + (p.z - Lcell);			
			
			avg_rise_vel += UPz*dt;	
			num_part_tot_time += dt;
			N++;
#else			
			//Step 5: Verify BCs:
			escape = verify_boundary_conditions(0, NBN, &p, bp, t, l, &TOE_verify_bc, fp[7]);
			
			if(escape==1)	break;	//to next particle		
				
			//Record particle trajectory:
			if(N < NUM_RECORD)
			{
				time[N] = t;	r[N] = p.r;			z[N] = p.z;		
				lsE[N] = lE;	rE[N] = eddy.r;		zE[N] = eddy.z;	
				N++;
			}			
			//Sample Coefficients of every nth particle:
			if(l%10==0 && M <= 1e8)
			{
				stats[0][M] = dt;
				stats[1][M] = Rer;
				stats[2][M] = Rez;
				stats[3][M] = CDr;
				stats[4][M] = CDz;
				stats[5][M] = 0.5*rhoF*NORM( AreaPr*CDr*relUr*relUr, AreaPz*CDz*relUz*relUz );
				stats[6][M] = NORM(FrAM, FzAM);
				stats[7][M] = NORM(FrPG, FzPG);
				//FzBuoy = const
				M++;
			}

#endif	//CELLULAR	
				
		} //End of Time Loop
		TOE_time_loop += (clock() - begin_time_loop);	

#ifdef CELLULAR	
		//Print final particle position:	
		fprintf(fp[1],"%g %g\n", p.r, p.z);
#else		
		//Print Residence time and Trajectories of Particle and Eddies:
		find_ordinal(l+1, ordinal);
		if(escape)		
		{
			printf("%d%s particle escaped with Tres = %g\n", l+1, ordinal, t);
			Tres_esc[num_esc++] = t;	//for plotting
		}
		else	printf("%d%s particle never escaped!\n", l+1, ordinal);
		Tres[l] = t;	//For statistics
		
		//Printing Trajectories: [To see how far drops go in LP and NOT how fast they go back to UP]		
		if(l < MAX(100, NUM_PART/10))		//Limit for Vizualization (@ Gnuplot)
		{			
			printdt = MAX(1, (double)N/NUM_PRINT);
			for(j=0; j<NUM_PRINT; j++)
			{
				i = (int)(j*printdt);		//print at some interval			
				if(i >= N)	break;				
				fprintf(fp[5], "%g %g %g %g %g %g %g\n",\
						time[i], r[i], z[i], rE[i], zE[i], lsE[i], pow(RP*200, 0.5));
			}
			N--;	//print the (N-1)th record:
			fprintf(fp[5], "%g %g %g %g %g %g %g\n\n",\
						time[N], r[N], z[N], rE[N], zE[N], lsE[N], pow(RP*200, 0.5));		
		}
#endif	//CELLULAR				
	}//End of Particle Loop
	num_ensemble = M;
	TOE_particle_loop += (clock() - begin_particle_loop);		
	
	
	/***************************************************************************************/
	/*********************** Various Output Statisics **************************************/
	/***************************************************************************************/
#ifdef CELLULAR
	avg_rise_vel = avg_rise_vel/num_part_tot_time/prop->Ucell;
	fprintf(fp[2],"%g\n", avg_rise_vel);
	printf("Dimensionless Average Rise Velocity: %g\n\n", avg_rise_vel);
#else		
	
	/******Time-step-weighted Mean and Coefficient of Variation of Various Coefficients*****/
	double mean[num_out_stats], cov[num_out_stats], tot_time = 0;
	
	for(i=0; i<num_ensemble; i++)	
	tot_time += stats[0][i];		

	for(k=0; k<num_out_stats; k++)	//"dt" is number averaged
	{
		mean[k] = 0;
		for(i=0; i<num_ensemble; i++)
		mean[k] += stats[k][i] * (k==0 ? 1 : stats[0][i]);	//stats[0][i] = dt
		mean[k] = mean[k]/(k==0 ? num_ensemble : tot_time);
	}	
	for(k=0; k<num_out_stats; k++)
	{
		cov[k] = 0;
		for(i=0; i<num_ensemble; i++)
		cov[k] += pow(stats[k][i] - mean[k], 2) * (k==0 ? 1 : stats[0][i]);
		cov[k] = pow(cov[k]/(k==0 ? num_ensemble : tot_time), 0.5)/mean[k];
	}	
	fprintf(fp[6],"Eo: %g\nMo: %g\n", Eo, Mo);
	fprintf(fp[6],"Mean_and_Coefficient_of_Variation_of_Various_Coefficients:\n");
	fprintf(fp[6],"dt: %g %g\nRe_r: %g %g\nRe_z: %g %g\nCDr: %g %g\nCDz: %g %g\n"
		"FDrag: %g %g\nFAM: %g %g\nFPG: %g %g\nFBuoy: %g 0\nsumdz/sumdr: %g %g\n\n",\
		mean[0], cov[0], mean[1], cov[1], mean[2], cov[2], mean[3], cov[3],	mean[4], cov[4],\
		mean[5]/(mF-mP), cov[5], mean[6]/(mF-mP), cov[6], mean[7]/(mF-mP), cov[7], 
		FzBuoy/(mF-mP), sumdz/sumdr, NORM(sumdr, sumdz));
			
	//Monitoring Flutuating Velocities at (pr0, pz0) ie. at t = 0:
	printf("\n\nAt Location r= %0.4f; z= %0.4f\n", pr0, pz0);
	printf("with velocities: Ur= %0.4f; Uz= %0.4f\n", UrX, UzX);
	printf("and TKE= %0.4f;	√(2/3*TKE)= %0.4f\n", TKEX, urmsX);
	printf("psi_avg= %g; psi_min = %g; psi_max = %g\n",psi_avg/NUM_PART, psi_min, psi_max);
	

	/**********************Residence Time Statistics****************************************/
	
	N = 20;		//Total number of bins
	int num[N], id[100];
	double minTres, maxTres, avgTres = 0, SDTres = 0, bin[N];
	
	//******Of Escaped particles - only for plotting RTD of limited bins:	
	struct qsvmc v[NUM_PART];
	for(l=0; l<num_esc; l++)
	v[l].s[0] = Tres_esc[l];	
	quick_sort (0, v, num_esc);	//sort ascending
	
	maxTres = v[num_esc-1].s[0];//Upper limit for RTD
	minTres = 0.1;				//Lower limt
	
	for(i=0; i<=N; i++)			//form bins
	{
		//bin[i] = (maxTres - 0)/N*i;
		bin[i] = minTres*pow(maxTres/minTres, (double)(i)/N);	//logarithmic
		num[i] = 0;
	}
	for(l=0; l<num_esc; l++)	//then sort
	for(i=0; i<N; i++)
	{
		if(Tres[l] > bin[i] && Tres[l] <= bin[i+1])
		num[i+1]++;
	}
	
	//******Of All particles:	
	for(l=0; l<NUM_PART; l++)
	avgTres += Tres[l];
	avgTres = avgTres/NUM_PART;
	
	for(l=0; l<NUM_PART; l++)
	SDTres += pow(Tres[l] - avgTres, 2);
	SDTres = pow(SDTres, 0.5);
	
	fprintf(fp[6],"avg_maxTres_esc: %g\navgTres_all: %g\nSDTres_all: %g\n"
					"Escaped_Particles: %d\n\n", maxTres, avgTres, SDTres, num_esc);
	
	for(i=0; i<N; i++)
	fprintf(fp[6],"%0.2f %d\n", bin[i+1], num[i+1]);
	/***********************************************************************************/
	
	//Time of Execution Statistics:
	TOE_particle_loop = TOE_particle_loop/CLOCKS_PER_SEC;	
	TOE_time_loop = TOE_time_loop/CLOCKS_PER_SEC;	
	TOE_interpolate = TOE_interpolate/CLOCKS_PER_SEC;
	TOE_verify_bc = TOE_verify_bc/CLOCKS_PER_SEC;
	printf("\nTime of Execution Statistics:\n"
			"Particle Tracking particle loop:	%g\n"
			"Particle Tracking time loop:		%g (%0.0f pct)\n"
			"Interpolation (eddy & particle):	%g (%0.0f pct)\n"
			"Verify Boundary Conditions:		%g (%0.0f pct)\n",\
			TOE_particle_loop, TOE_time_loop, TOE_time_loop/TOE_particle_loop*100,\
			TOE_interpolate, TOE_interpolate/TOE_particle_loop*100,\
			TOE_verify_bc, TOE_verify_bc/TOE_particle_loop*100);
#endif	//CELLULAR			

	free (r);
	free (z);
	free (rE);
	free (zE);
	free (lsE);
	free (time);
	for (i=0; i<num_out_stats; i++) 
	free(stats[i]);
	return;
}

/***************************************************************************************/
/***********************************VARIOUS FUNCTIONS***********************************/
/***************************************************************************************/

//http://stackoverflow.com/questions/849211/shortest-distance-between-a-point-and-a-line-segment
double edge_point_distance (double vx, double vy, double wx, double wy, double px, double py)
{
	// Find minimum distance between line segment v-w and point p
	double vw2 = (vx-wx)*(vx-wx) + (vy-wy)*(vy-wy);	// |v-w|^2
	if(vw2 == 0) 	return vw2;   					// v = w case
	
	// Consider the line extending the segment, parameterized as v + t (w - v).
	// We find projection of point p onto the line. 
	// It falls where t = [(p-v) . (w-v)] / |w-v|^2	
	double t = ((px - vx)*(wx - vx) + (py - vy)*(wy - vy))/vw2;

	if (t < 0.0)	return NORM(px-vx, py-vy);	// Beyond the 'v' end of the segment
	if (t > 1.0)	return NORM(px-wx, py-wy);  // Beyond the 'w' end of the segment
	
	double proj_x = vx + t*(wx - vx);
	double proj_y = vy + t*(wy - vy);
	return NORM(px - proj_x, py - proj_y);  	// Projection falls on the segment
}

int verify_boundary_conditions(int eddy, int NBN, struct particle *p, \
			 struct bpv *bp, double time, int particle_id, double *TOE, FILE *fp)		
{
	clock_t begin_TOE = clock();	
	
	int i, bpid, sign1, sign2, rand_sign, escape = 0;
	double r1, z1, r2, z2, dist1, dist2 = 123456789;
	double rIV, zIV, rBPIV, zBPIV, normBPIV, Nr_dist1, Nr_dist2, del;
	
	//Find the nearest boundary edge (NBE) to p:
	for(i=0; i<NBN; i++)	//in the lot of NBN (not NBN-1) edges
	{
		r1 = bp[i].r;		z1 = bp[i].z;		//bp1	
		r2 = bp[i+1].r;		z2 = bp[i+1].z;		//bp2
		
		/*/Distance from p to mid point of edge bp1-bp2 - Not for Reduced bp's
		dist1 = NORM( ((r1 + r2)/2 - p->r), ((z1 + z2)/2 - p->z) );*/
		
		//Nearest distance from p to edge bp1-bp2: For reduced bp's
		dist1 = edge_point_distance (r1, z1, r2, z2, p->r, p->z);
		if(dist2 > dist1)	
		{ dist2 = dist1;	bpid = i; }				
	}	
	r1 = bp[ bpid ].r;		z1 = bp[ bpid ].z;		//bp1	
	r2 = bp[ bpid+1 ].r;	z2 = bp[ bpid+1 ].z;	//bp2
	//Point of inward vector (pIV) of bp1:
	rIV = bp[ bpid ].rIV;	zIV = bp[ bpid ].zIV;	
	//Numerator of Distance from edge bp1-bp2 to p: 
	Nr_dist1 = (r2-r1)*(z2 - p->z) - (z2-z1)*(r2 - p->r);					
	//Numerator of Distance from edge bp1-bp2 to pIV:	
	Nr_dist2 = (r2-r1)*(z2 - zIV) - (z2-z1)*(r2 - rIV);

	/*if(particle_id==92 && eddy==0)
	fprintf(fp, "%g %g %g %g %g %g %g %g %g, %d\n", 
			time, p->r, p->z, r1, z1, r2, z2, rIV, zIV, bpid);*/
	
	//If 'p' is outside of 'NBE'...:
	if(SIGN(Nr_dist1) != SIGN(Nr_dist2))
	{
		if(eddy)	//If it was an eddy thrown across a boundary
		escape = 1;	//	new eddy will be assigned

		//If particle crossed the slag-metal interface:
		else if( bp[ bpid ].BcType == 2 && bp[ bpid+1 ].BcType == 2 )
		escape = 1;	//particle escapes
			
		else	//Else limit its trajectory within 'del' from boundary edge:		
		{
			del = (bp[ bpid ].BcType==3 ? 5*dL : dL);
			//del is larger for BcType=3 to avoid drops sticking in recesses
			
			dist1 = fabs(Nr_dist1)/NORM(r2-r1, z2-z1);
			rBPIV = rIV - r1;	
			zBPIV = zIV - z1;
			normBPIV = NORM(rBPIV, zBPIV);
			
			//Avoid (rBPIV, zBPIV) = 0 by slightly rotating the vector:
			// "0.1*normBPIV" will give a rotation of ~6° to the vector
			rand_sign = (RANDNUM() > 0 ? 1 : -1);
			rBPIV = (rBPIV > 0 ? MAX(rBPIV, 0.1*normBPIV) : \
					(rBPIV < 0 ? MIN(rBPIV, -0.1*normBPIV) : rand_sign*0.1*normBPIV));
			zBPIV = (zBPIV > 0 ? MAX(zBPIV, 0.1*normBPIV) : \
					(zBPIV < 0 ? MIN(zBPIV, -0.1*normBPIV) : rand_sign*0.1*normBPIV));			
			
			p->r = p->r + (dist1 + del)*rBPIV/normBPIV;
			p->z = p->z + (dist1 + del)*zBPIV/normBPIV;	
			escape = 2;
		}
	}
	*TOE += clock() - begin_TOE;
	return escape;
}

		
void interpolate_scalars(int eddy, int NC, int NR, double *rdist, double *zdist,
		struct nodal **node, struct particle *p, double time, int particle_id, 
		double *TOE, FILE *fp )
{
	clock_t begin_TOE = clock();
	
	struct nodal *n[4];
	int i, j, idrow, idcol, enid[4], nnid, spout = 0;
	double rNN, zNN, dist1, dist2, pr = p->r, pz = p->z, r[3], z[3], S[20][4], val[20];
	
	//Step-1: Find Nearest r-column z-row
	dist1 = 123456789;
	for(j=1; j<NC-1; j++)
	{
		dist2 = fabs(pr - rdist[j]);
		if(dist1 > dist2)	
		{ dist1 = dist2;	idcol = j; }
	}

	dist1 = 123456789;
	for(j=1; j<NR-1; j++)
	{
		dist2 = fabs(pz - zdist[j]);
		if(dist1 > dist2)	
		{ dist1 = dist2;	idrow = j; }		
	}
	//(r,z) of N-earest N-ode:
	rNN = rdist[idcol];
	zNN = zdist[idrow];

	//Step-2: Find four surrounding Nodes
	if( pr <= rNN && pz <= zNN )			// < <
	{
		n[3] = &node[ idcol ][ idrow ];
		n[1] = &node[ idcol ][ idrow-1 ];
		n[0] = &node[ idcol-1 ][ idrow-1 ];
		n[2] = &node[ idcol-1 ][ idrow ];
	}
	else if( pr >= rNN && pz <= zNN )		// > <
	{
		n[2] = &node[ idcol ][ idrow ];
		n[3] = &node[ idcol+1 ][ idrow ];
		n[1] = &node[ idcol+1 ][ idrow-1 ];
		n[0] = &node[ idcol ][ idrow-1 ];
	}		
	else if( pr <= rNN && pz >= zNN )		// < >
	{	
		n[1] = &node[ idcol ][ idrow ];
		n[0] = &node[ idcol-1 ][ idrow ];
		n[2] = &node[ idcol-1 ][ idrow+1 ];
		n[3] = &node[ idcol ][ idrow+1 ];
	}
	else if( pr >= rNN && pz >= zNN )		// > >
	{
		n[0] = &node[ idcol ][ idrow ];
		n[2] = &node[ idcol ][ idrow+1 ];
		n[3] = &node[ idcol+1 ][ idrow+1 ];
		n[1] = &node[ idcol+1 ][ idrow ];
	}
	/*Four nodes must be ordered as (r1,z1), (r2,z1), (r1,z2), (r2,z2): 

	RHS: For clarity, nearest node (node[ idcol ][ idrow ]) is first written
		and then rest of the nodes written w.r.t this node, in clockwise order
		(Nearest Node is at the intersection of nearest col and nearest row)
														2	3
	LHS: Ids to nodes (n[id]) are given in this order:
														0	1
	Since r, z values are originally filled in ascending order, 
	idXXX + 1 (in node[][]) would give give increasing r, z values and vice versa*/
	
	
	/*if(particle_id>=9 && time > 18.1 && eddy) {
		for(i=0; i<4; i++)
		printf("%g %g %g %g\n",n[i]->r, n[i]->z, n[i]->Ur, n[i]->Uz);
		printf("%g %g %g %g %d %d\n\n",pr, pz, rNN, zNN, idcol, idrow); }*/
	
	//Step-3: Collect scalars of nodes and Send for Interpolation
	for(i=0; i<4; i++)
	{
		//S[][i=0,1,2,3] <=> S[][11, 21, 12, 22] 
		S[0][i] = n[i]->Ur;
		S[1][i] = n[i]->Uz;	
		if(eddy) continue;
		S[2][i] = n[i]->Ar;
		S[3][i] = n[i]->Az;
		S[4][i] = n[i]->TKE;
		S[5][i] = n[i]->eps;	
		S[6][i] = n[i]->dUrdr;
		S[7][i] = n[i]->dUzdr;
		S[8][i] = n[i]->dUrdz;
		S[9][i] = n[i]->dUzdz;		
	}

	r[0] = pr;			z[0] = pz;		
	r[1] = n[0]->r;		z[1] = n[1]->z;		
	r[2] = n[1]->r;		z[2] = n[2]->z;	

	if(eddy)	bilinear_interpolation(2, r, z, S, val);
	else 		bilinear_interpolation(10, r, z, S, val);
	p->Ur	= val[0];
	p->Uz	= val[1]; 
	 
	if(!eddy)
	{
		p->Ar	= val[2];
		p->Az	= val[3];
		p->TKE	= val[4];
		p->eps	= val[5];
		p->dUrdr = val[6];
		p->dUzdr = val[7];
		p->dUrdz = val[8];
		p->dUzdz = val[9];		

		//If particle is in the spout region...
		p->spoutBc = ( n[0]->BcType || n[1]->BcType || n[2]->BcType || n[3]->BcType );
	}	
	*TOE += clock() - begin_TOE;
	return;	
}

void bilinear_interpolation(int N, double *r, double *z, double S[N][4], double *val)
{
	int i;
	double R = r[0], r1 = r[1], r2 = r[2], Z = z[0], z1 = z[1], z2 = z[2];

	for(i=0; i<N; i++)
	val[i] = 1.0/((r2-r1)*(z2-z1))*\
				( S[i][0]*(r2-R)*(z2-Z) + S[i][1]*(R-r1)*(z2-Z) +\
				  S[i][2]*(r2-R)*(Z-z1) + S[i][3]*(R-r1)*(Z-z1) );		
	return;
}

double angle_bet_vectors(struct bpv *p)
{
	//cos(theta) = a.b/(|a||b|); a = (p1-p0); b = (p3-p2);
	double Nr, Dr;
	Nr = (p[1].r - p[0].r)*(p[3].r - p[2].r) + (p[1].z - p[0].z)*(p[3].z - p[2].z);
	Dr = NORM((p[1].r - p[0].r), (p[1].z - p[0].z))*NORM((p[3].r - p[2].r), (p[3].z - p[2].z));
	return 180/PI*acos(Nr/Dr);
}
void reduce_boundary_edges (int *NBN0, struct bpv *bp, double angle_spec, double interval)
{
	int i, last_i, M = 0, NBN = *NBN0, bpid[*NBN0];
	double angle;
	struct bpv p[4];
	
	bpid[M++] = 0;
	bpid[M++] = 1;
	for(i=2; i<NBN; i++)
	{		
		if(i+1 >= NBN)	break;
		
		if( bp[i].BcType != bp[i+1].BcType )	//When BcType changes w.r.t next node
		{
			bpid[M++] = i;
			bpid[M++] = i+1;
		}	//No nodes can be skipped here as they are required to check dir. change
			
		p[0] = bp[i-1];	
		p[1] = p[2] = bp[i];		
		p[3] = bp[i+1];		
		angle = angle_bet_vectors(p);
		
		if(angle > angle_spec)	//If there is a significant turn
		{							
			//count three consecutive points if not counted earlier
			if(i != bpid[M-2] && i != bpid[M-1])
			bpid[M++] = i;
			if(i+1 != bpid[M-1])
			bpid[M++] = i+1;
			last_i = bpid[M++] = i+2;
		}
		else if(angle > 0.1)	//i.e. non-zero, small turns
		{
			//account once in "interval" points
			if(i != bpid[M-1] && i > last_i + interval)
			last_i = bpid[M++] = i;	
		}		
	}
	if(i != bpid[M-1])	bpid[M++] = i;	//account the penultimate bp
	
	*NBN0 = M ;
	for(i=0; i<M; i++)
	bp[i] = bp[ bpid[i] ];
	bp[M] = bp[0];		//connecting first and last bp
	return;
}
	
void concatenate_boundary_points (int NBN, struct bpv *bp)
{
	int i, j, N = NBN, pid;
	double dist1, dist2;
	struct bpv p[NBN];

	N--;		
	for(i=0; i<N; i++)
	p[i] = bp[i+1];		//copy bp into p excluding bp[0]

	for(i=1; i<NBN; i++)
	{
		dist2 = 123456789;
		for(j=0; j<N; j++)	
		{		
			dist1 = NORM( (bp[i-1].r - p[j].r), (bp[i-1].z - p[j].z) );
			if(dist2 > dist1)	
			{ dist2 = dist1; 	pid = j; }
		}
		bp[i] = p[pid];
		//Remove p[pid]:
		N--;		
		for(j=pid; j<N; j++)	
		p[j] = p[j+1];	//pull the array up				
	}
	bp[NBN] = bp[0];	//connecting last and first bp
	return;
}

void quick_sort_multi_column ( struct qsvmc *v, int N,  int M )
{
	//First sorting:
	quick_sort (0, v, N);
	
	//The rest:
	int i,j,n;
	double v0;
	for(i=1; i<M; i++)	//sorting recursively up to M number of columns
	{
		n = 0;
		v0 = v[0].s[i-1];	
		for(j=0; j<N; j++)
		{		
			if( fabs(v[j].s[i-1] - v0) > 1e-5 ) 	// v_ji != v0
			{
				quick_sort (i, &v[j-n], n);
				v0 = v[j].s[i-1];
				j--;
				n = 0;
			}
			else	n++;
		}
		quick_sort (i, &v[j-n], n);
	}
}
//Adapted from http://rosettacode.org/wiki/Sorting_algorithms/Quicksort#C
void quick_sort (int var_id, struct qsvmc *v, int n) 
{
	if (n < 2)
	return;
	struct qsvmc temp, *LEFT, *RIGHT;
	double pivot = v[n/2].s[var_id];	
	LEFT = v;			// address of LEFT most element
	RIGHT = v + n - 1;	// address of RIGHT most element
	while (LEFT <= RIGHT) 			
	{
		while (LEFT->s[var_id] < pivot)		LEFT++;
		while (RIGHT->s[var_id] > pivot)	RIGHT--;
		if (LEFT <= RIGHT) 
		{
			 temp = *LEFT;
			*LEFT++ = *RIGHT;
			*RIGHT-- =  temp;
		}
	}
	quick_sort(var_id, v, RIGHT-v+1);
	quick_sort(var_id, LEFT, v+n-LEFT);	
}

void dimensionalize_vel_accel(int choice, int NC, int NR, 
									double Ucell, struct nodal **node)
{	/*******For cellular flow verification case*******/
	int j, k;
	struct nodal *nC;	
	double F = (choice==1 ? Ucell : 1.0/Ucell);
	
	for(j=0; j<NC; j++)
	for(k=0; k<NR; k++)
	{		
		if(!node[j][k].exist) continue;
		nC = &node[j][k];
		nC->Ur = nC->Ur*F; 
		nC->Uz = nC->Uz*F;			
		nC->Ar = nC->Ar*F*F; 
		nC->Az = nC->Az*F*F; 
		nC->dUrdr = nC->dUrdr*F;
		nC->dUzdr = nC->dUzdr*F;
		nC->dUrdz = nC->dUrdz*F;
		nC->dUzdz = nC->dUzdz*F;
		
	}
}
